from __future__ import annotations

__all__ = [
    "align_categories",
    "is_identifier",
    "reindex_series",
    "sort_nicely",
    "sum_df_sequence",
]

import re
from collections.abc import Sequence
from keyword import kwlist
from typing import Dict, List, Union

from pandas import DataFrame, Index, MultiIndex, Series


def reindex_series(
    series: Series,
    target_series: Series,
    *,
    source_levels: List[int] = None,
    target_levels: List[int] = None,
    fill_value: Union[int, float] = None,
) -> Series:
    # Make shallow copies of the source and target series in case their indexes need to be changed
    series = series.copy(deep=False)
    target_series = target_series.copy(deep=False)

    if series.index.nlevels > 1 and source_levels is not None:
        arrays = [series.index.get_level_values(level) for level in source_levels]
        series.index = MultiIndex.from_arrays(arrays)

    if target_series.index.nlevels > 1 and target_levels is not None:
        arrays = [target_series.index.get_level_values(level) for level in target_levels]
        target_series.index = MultiIndex.from_arrays(arrays)

    reindexed = series.reindex(target_series.values, fill_value=fill_value)
    reindexed.index = target_series.index

    return reindexed


def align_categories(
    items: Sequence[Union[Series, DataFrame]],
):
    """Pre-processing step for ``pd.concat()`` which attempts to align any Categorical series in the sequence to using
    the same set of categories. It passes through the sequence twice: once to accumulate the complete set of all
    categories used in the sequence; and a second time to modify the sequence's contents to use this full set. The
    contents of the sequence are modified in-place.

    Note:
        The resulting categories will be lex-sorted (based on the ``sorted()`` builtin)

    Args:
        items (Sequence[pandas.Series | pandas.DataFrame]): Any sequence of Series or DataFrame objects (anything that
            is acceptable to ``pandas.concat()``)
    """
    item_type = None
    for item in items:
        if item_type is None:
            if isinstance(item, DataFrame):
                item_type = DataFrame
            elif isinstance(item, Series):
                item_type = Series
            else:
                raise TypeError(type(item))
        else:
            assert isinstance(item, item_type)

    if item_type is Series:
        _align_series_categories(items)
    else:
        column_categories = _enumerate_frame_categories(items)
        _align_frame_categories(items, column_categories)


def _align_series_categories(series_list: Sequence[Series]):
    all_categories = set()
    for series in series_list:
        if not hasattr(series, "cat"):
            raise TypeError()
        all_categories |= set(series.cat.categories)

    sorted_categories = sorted(all_categories)
    for series in series_list:
        missing_categories = all_categories.difference(series.cat.categories)
        if missing_categories:
            series.cat.add_categories(missing_categories, inplace=True)
        series.cat.reorder_categories(sorted_categories, inplace=True)


def _enumerate_frame_categories(frames: Sequence[DataFrame]) -> Dict[str, set]:
    column_categories = {}
    for frame in frames:
        for col_name, series in frame.items():
            if not hasattr(series, "cat"):
                continue
            categories = set(series.cat.categories)

            if col_name not in column_categories:
                column_categories[col_name] = categories
            else:
                column_categories[col_name] |= categories
    return column_categories


def _align_frame_categories(frames: Sequence[DataFrame], column_categories: Dict[str, set]):
    for col_name, all_categories in column_categories.items():
        sorted_categories = sorted(all_categories)
        for frame in frames:
            if col_name not in frame:
                continue
            s = frame[col_name]
            missing_categories = all_categories.difference(s.cat.categories)
            if missing_categories:
                s.cat.add_categories(missing_categories, inplace=True)
            s.cat.reorder_categories(sorted_categories, inplace=True)


def sum_df_sequence(
    items: Sequence[DataFrame],
    *,
    fill_value: Union[int, float] = 0,
) -> DataFrame:
    """Sums over a sequence of DataFrames, even if they have different indexes or columns, filling in 0 (or a value of
    your choice) for missing rows or columns. Useful when you have a sequence of DataFrames which are supposed to have
    the same indexes and columns but might be missing a few values.

    Args:
        items (Sequence[pandas.DataFrame]): Any sequence of DataFrames, ordered or unordered.
        fill_value (int | float, optional): Defaults to ``0``. The value to use for missing cells.

    Returns:
        pandas.DataFrame: The sum over all items in seq.
    """
    common_index = Index([])
    common_columns = Index([])
    accumulator = DataFrame()

    for df in items:
        if not df.index.equals(common_index):
            common_index |= df.index
            accumulator = accumulator.reindex(common_index, axis=0, fill_value=fill_value)
            df = df.reindex(common_index, axis=0, fill_value=fill_value)
        if not df.columns.equals(common_columns):
            common_columns |= df.columns
            accumulator = accumulator.reindex(common_columns, axis=1, fill_value=fill_value)
            df = df.reindex(common_columns, axis=1, fill_value=fill_value)
        accumulator += df
    return accumulator


def is_identifier(name: str) -> bool:
    """Tests that the name is a valid Python variable name and does not collide with reserved keywords

    Args:
        name (str): Name to test

    Returns:
        bool: If the name is 'Pythonic'
    """

    return name.isidentifier() and name not in kwlist


def _tryint(s):
    try:
        return int(s)
    except ValueError:
        return s


def _alphanum_key(s):
    """Turn a string into a list of string and number chunks (eg. "z23a" -> ["z", 23, "a"])"""
    return [_tryint(c) for c in re.split("([0-9]+)", s)]


def sort_nicely(
    l: List[str],
) -> List[str]:
    """Sort the given list of strings in the way that humans expect.

    Args:
        l (List[str]): List of strings to sort.

    Returns:
        List[str]: The sorted list of strings
    """
    l = l.copy()
    l.sort(key=_alphanum_key)
    return l



from __future__ import annotations

__all__ = [
    "generate_time_bin_labels",
]

from datetime import datetime, timedelta
from typing import Dict


def generate_time_bin_labels(
    day_start_minute: int,
    day_end_minute: int,
    *,
    time_resolution_minute: int = 30,
    base_dt: datetime = None,
) -> Dict[int, datetime]:
    """Generate time bin labels based on start time, end time, and time resolution.

    Args:
        day_start_minute (int): Start time of day, in minutes past midnight
        day_end_minute (int): End time of day, in minutes past midnight
        time_resolution_minute (int, optional): Defaults to ``30``. The number of minutes in each time bin
        base_dt (datetime, optional): Defaults to ``None``. A datetime object to use as a base for generating the time
            bin labels. All properties of the provided datetime object will be inherited except hour, minute, second,
            and microsecond. If ``None``, the current date will be used.

    Returns:
        dict[int, datetime]: A dictionary with the time bins as its keys and datetime objects as its values
    """

    if day_start_minute > day_end_minute:
        raise ValueError("`day_end_minute` must be > `day_start_minute`")

    diff = day_end_minute - day_start_minute
    if diff < time_resolution_minute:
        raise ValueError("`day_start_minute - day_end_minute` must be > `time_resolution_minute`")

    start_hr, start_min = divmod(day_start_minute, 60)
    n_intervals = diff // time_resolution_minute

    if base_dt is None:
        base_dt = datetime.today()
    start_time = base_dt.replace(hour=start_hr, minute=start_min, second=0, microsecond=0)
    labels = {(i + 1): start_time + timedelta(seconds=time_resolution_minute * 60 * i) for i in range(n_intervals)}

    return labels



from __future__ import annotations

__all__ = [
    "peek_mdf",
    "read_emx",
    "read_mdf",
    "to_emx",
    "to_mdf",
]

from io import FileIO
from os import PathLike
from typing import Iterable, List, Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from .common import coerce_matrix, open_file

# region EMME Matrices


def read_mdf(
    file: Union[str, FileIO, PathLike],
    *,
    raw: bool = False,
    tall: bool = False,
) -> Union[NDArray, pd.DataFrame, pd.Series]:
    """Reads Emme's official matrix "binary serialization" format, created using ``inro.emme.matrix.MatrixData.save()``.
    There is no official extension for this type of file; '.mdf' is recommended. '.emxd' is also sometimes encountered.

    Args:
        file (str | FileIO | PathLike): The file to read.
        raw (bool, optional): Defaults to ``False``. If ``True``, returns an unlabelled ndarray. Otherwise, a DataFrame
            will be returned.
        tall (bool, optional): Defaults to ``False``. If ``True``, a 1D data structure will be returned. If
            ``raw=False``, a Series will be returned, otherwise a 1D ndarray.

    Returns:
        NDArray, pandas.DataFrame, or pandas.Series: The matrix stored in the file.
    """
    with open_file(file, mode="rb") as file_handler:
        magic, version, dtype_index, ndim = np.fromfile(file_handler, np.uint32, count=4)

        if magic != 0xC4D4F1B2 or version != 1 or not (0 < dtype_index <= 4) or not (0 < ndim <= 2):
            raise IOError(
                "Unexpected file header: magic number: %X, version: %d, data type: %d, dimensions: %d."
                % (magic, version, dtype_index, ndim)
            )

        shape = np.fromfile(file_handler, np.uint32, count=ndim)

        index_list = []
        for n_items in shape:
            indices = np.fromfile(file_handler, np.int32, n_items)
            index_list.append(indices)

        dtype = {1: np.float32, 2: np.float64, 3: np.int32, 4: np.uint32}[dtype_index]
        flat_length = shape.prod()  # Multiply the shape tuple
        matrix = np.fromfile(file_handler, dtype, count=flat_length)

        if raw and tall:
            return matrix

        matrix.shape = shape

        if raw:
            return matrix

        if ndim == 1:
            return pd.Series(matrix, index=index_list[0])
        elif ndim == 2:
            matrix = pd.DataFrame(matrix, index=index_list[0], columns=index_list[1])

            return matrix.stack() if tall else matrix

        raise NotImplementedError()  # This should never happen


def to_mdf(
    matrix: Union[pd.DataFrame, pd.Series],
    file: Union[str, FileIO, PathLike],
) -> None:
    """Writes a matrix to Emme's official "binary serialization" format, which can be loaded in Emme using
    ``inro.emme.matrix.MatrixData.load()``. There is no official extension for this type of file; '.mdf' is recommended.

    Args:
        matrix (pandas.DataFrame | panda.Series): The matrix to write to disk. If a Series is given, it MUST have
            a MultiIndex with exactly 2 levels to unstack.
        file (str | FileIO | PathLike): The path or file handler to write to.
    """
    if isinstance(matrix, pd.Series):
        row_index = matrix.index.get_level_values(0).unique()
        column_index = matrix.index.get_level_values(1).unique()
    elif isinstance(matrix, pd.DataFrame):
        row_index = matrix.index
        column_index = matrix.columns
    else:
        raise TypeError("Only labelled matrix objects are supported")

    with open_file(file, mode="wb") as writer:
        data = coerce_matrix(matrix, allow_raw=False)

        np.array([0xC4D4F1B2, 1, 1, 2], dtype=np.uint32).tofile(writer)  # Header
        np.array(data.shape, dtype=np.uint32).tofile(writer)  # Shape

        np.array(row_index, dtype=np.int32).tofile(writer)
        np.array(column_index, dtype=np.int32).tofile(writer)

        data.tofile(writer)


def peek_mdf(
    file: Union[str, FileIO, PathLike],
    *,
    as_index: bool = True,
) -> Union[List[List[int]], List[pd.Index]]:
    """Partially opens an MDF file to get the zone system of its rows and its columns.

    Args:
        file (str | FileIO | PathLike): The file to read.
        as_index (bool, optional): Defaults to ``True``. Set to ``True`` to return a pandas.Index object rather than
            List[int]

    Returns:
        List[int] or List[pandas.Index]: One item for each dimension. If ``as_index=True``, the items will be pandas.Index objects, otherwise they will be List[int]
    """
    with open_file(file, mode="rb") as file_handler:
        magic, version, dtype_index, ndim = np.fromfile(file_handler, np.uint32, count=4)

        if magic != 0xC4D4F1B2 or version != 1 or not (0 < dtype_index <= 4) or not (0 < ndim <= 2):
            raise IOError(
                "Unexpected file header: magic number: %X, version: %d, data type: %d, dimensions: %d."
                % (magic, version, dtype_index, ndim)
            )

        shape = np.fromfile(file_handler, np.uint32, count=ndim)

        index_list = []
        for n_items in shape:
            indices = np.fromfile(file_handler, np.int32, n_items)
            index_list.append(indices)

        if not as_index:
            return index_list

        return [pd.Index(zones) for zones in index_list]


def read_emx(
    file: Union[str, FileIO, PathLike],
    *,
    zones: Union[int, Iterable[int], pd.Index] = None,
    tall: bool = False,
) -> Union[NDArray, pd.DataFrame, pd.Series]:
    """Reads an "internal" Emme matrix (found in `<Emme Project>/Database/emmemat`); with an '.emx' extension. This data
    format does not contain information about zones. Its size is determined by the dimensions of the Emmebank
    (``Emmebank.dimensions['centroids']``), regardless of the number of zones actually used in all scenarios.

    Args:
        file (str | File | PathLike): The file to read.
        zones (int | Iterable[int] | pandas.Index, optional): Defaults to ``None``. An Index or Iterable will be
            interpreted as the zone labels for the matrix rows and columns; returning a DataFrame or Series (depending
            on ``tall``). If an integer is provided, the returned ndarray will be truncated to this 'number of zones'.
            Otherwise, the returned ndarray will be size to the maximum number of zone dimensioned by the Emmebank.
        tall (bool, optional): Defaults to ``False``. If True, a 1D data structure will be returned. If ``zone_index``
            is provided, a Series will be returned, otherwise a 1D ndarray.

    Returns:
        NDArray, pandas.DataFrame, or pandas.Series.

    Examples:
        For a project with 20 zones:

        >>> matrix = read_emx("Database/emmemat/mf1.emx")
        >>> print type(matrix), matrix.shape
        (numpy.ndarray, (20, 20))

        >>> matrix = read_emx("Database/emmemat/mf1.emx", zones=10)
        >>> print type(matrix), matrix.shape
        (numpy.ndarray, (10, 10))

        >>> matrix = read_emx("Database/emmemat/mf1.emx", zones=range(10))
        >>> print type(matrix), matrix.shape
        <class 'pandas.core.frame.DataFrame'> (10, 10)

        >>> matrix = read_emx("Database/emmemat/mf1.emx", zones=range(10), tall=True)
        >>> print type(matrix), matrix.shape
        <class 'pandas.core.series.Series'> 100

    """
    with open_file(file, mode="rb") as reader:
        data = np.fromfile(reader, dtype=np.float32)

        n = int(len(data) ** 0.5)
        if len(data) != n**2:
            raise RuntimeError("Matrix shape from file is not square")

        if zones is None and tall:
            return data

        data.shape = n, n

        if isinstance(zones, (int, np.int_)):
            data = data[:zones, :zones]

            if tall:
                data.shape = zones * zones
                return data
            return data
        elif zones is None:
            return data

        zones = pd.Index(zones)
        n = len(zones)
        data = data[:n, :n]

        matrix = pd.DataFrame(data, index=zones, columns=zones)

        return matrix.stack() if tall else matrix


def to_emx(
    matrix: Union[pd.DataFrame, pd.Series, NDArray],
    file: Union[str, FileIO, PathLike],
    emmebank_zones: int,
) -> None:
    """Writes an "internal" Emme matrix (found in `<Emme Project>/Database/emmemat`); with an '.emx' extension. The
    number of zones that the Emmebank is dimensioned for must be known in order for the file to be written correctly.

    Args:
        matrix (pandas.DataFrame | pandas.Series | numpy.ndarray): The matrix to write to disk. If a Series is
            given, it MUST have a MultiIndex with exactly 2 levels to unstack.
        file (str | FileIO | PathLike): The path or file handler to write to.
        emmebank_zones (int): The number of zones the target Emmebank is dimensioned for.
    """
    if emmebank_zones <= 0:
        raise ValueError("Emmebank zones must be > 0")

    with open_file(file, mode="wb") as writer:
        data = coerce_matrix(matrix)
        n = data.shape[0]
        if n > emmebank_zones:
            out = data[:emmebank_zones, :emmebank_zones].astype(np.float32)
        else:
            out = np.zeros([emmebank_zones, emmebank_zones], dtype=np.float32)
            out[:n, :n] = data

        out.tofile(writer)


# endregion



__all__ = [
    "read_feat",
    "read_table_calculator_result_attributes",
    "read_choice_model_component_tables",
    "read_utility_expression_table",
    "read_calibration_target_table",
    "coerce_matrix",
    "expand_array",
    "open_file",
    "read_fortran_rectangle",
    "read_fortran_square",
    "to_fortran",
    "peek_mdf",
    "read_emx",
    "read_mdf",
    "to_emx",
    "to_mdf",
    "read_nwp_base_network",
    "read_nwp_exatts_list",
    "read_nwp_link_attributes",
    "read_nwp_node_attributes",
    "read_nwp_traffic_results",
    "read_nwp_traffic_results_at_countpost",
    "read_nwp_transit_line_attributes",
    "read_nwp_transit_network",
    "read_nwp_transit_result_summary",
    "read_nwp_transit_segment_results",
    "read_nwp_transit_station_results",
    "read_nwp_transit_vehicles",
    "read_omx",
    "to_omx",
]

from .agent import (
    read_calibration_target_table,
    read_choice_model_component_tables,
    read_feat,
    read_table_calculator_result_attributes,
    read_utility_expression_table,
)
from .common import coerce_matrix, expand_array, open_file
from .fortran import read_fortran_rectangle, read_fortran_square, to_fortran
from .inro import peek_mdf, read_emx, read_mdf, to_emx, to_mdf
from .nwp import (
    read_nwp_base_network,
    read_nwp_exatts_list,
    read_nwp_link_attributes,
    read_nwp_node_attributes,
    read_nwp_traffic_results,
    read_nwp_traffic_results_at_countpost,
    read_nwp_transit_line_attributes,
    read_nwp_transit_network,
    read_nwp_transit_result_summary,
    read_nwp_transit_segment_results,
    read_nwp_transit_station_results,
    read_nwp_transit_vehicles,
)
from .omx import read_omx, to_omx



import unittest

import numpy as np
import pandas as pd
from pandas import testing as pdt

from wsp_balsa.routines.best_intermediates import (
    best_intermediate_subset_zones, best_intermediate_zones)


class TestBestIntermediates(unittest.TestCase):

    def setUp(self):
        normal_zones = [1, 2]
        intermediate_zones = [100, 101, 102, 103]

        self.pk_index = pd.MultiIndex.from_product([normal_zones, intermediate_zones], names=['p', 'k'])
        self.kq_index = pd.MultiIndex.from_product([intermediate_zones, normal_zones], names=['k', 'q'])
        self.pq_index = pd.MultiIndex.from_product([normal_zones, normal_zones], names=['p', 'q'])

        self.pk_distance = pd.Series(
            [5.093, 1.8829, 3.1976, 0.7403, 4.6856, 7.8176, 3.2228, 5.2814], index=self.pk_index
        )

        self.pk_table = pd.DataFrame({
            'util': [-0.1569, -0.4972, -0.4174, -0.1145, -0.1197, -0.0175, -0.3938, -0.9686],
            'col1': range(8),
            'col2': range(8)
        }, index=self.pk_index)

        self.kq_table = pd.DataFrame({
            'util': [-0.739, -np.inf, -np.inf, -0.3384, -0.9853, -0.1971, -np.inf, -0.896],
            'col1': range(8),
            'col3': range(8)
        }, index=self.kq_index)

    def test_best_path(self):
        expected_result = pd.DataFrame([
            [100, True, -0.8959, 0, 0, 0],
            [102, True, -0.6145, 7, 2, 5],
            [100, True, -0.8587, 4, 4, 0],
            [101, True, -0.3559, 8, 5, 3],
        ], index=self.pq_index, columns=['intermediate_zone', 'available', 'util', 'col1', 'col2', 'col3'])

        df = best_intermediate_zones(self.pk_table, self.kq_table, 'util')

        pdt.assert_frame_equal(expected_result, df)

    def test_best_subset_path(self):
        expected_result = pd.DataFrame([
            [100, True, -0.8959, 0, 0, 0],
            [101, True, -0.8356, 4, 1, 3],
            [100, True, -0.8587, 4, 4, 0],
            [102, True, -0.5909, 11, 6, 5],
        ], index=self.pq_index, columns=['intermediate_zone', 'available', 'util', 'col1', 'col2', 'col3'])

        df = best_intermediate_subset_zones(
            self.pk_distance, self.pk_table, self.kq_table, 'util', n_subset=2, maximize_subset=False
        )

        pdt.assert_frame_equal(expected_result, df)

    # TODO: More comprehensive tests


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

__all__ = [
    "best_intermediate_zones",
    "best_intermediate_subset_zones",
]

from multiprocessing import cpu_count
from threading import Thread
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray

try:
    from numba import njit
except ImportError:

    def njit(*args, **kwargs):
        def decorator(func):
            return func

        return decorator


_NULL_INDEX = -1
_NEG_INF = -np.inf


def _get_breaks(n_items: int, n_workers: int) -> List[Tuple[int, int]]:
    """Gets starts and stops for breaks similar to numpy.array_slice()"""
    div, remainder = divmod(n_items, n_workers)
    slices = []
    bottom = 0
    for i in range(n_workers):
        top = bottom + div
        top += int(i < remainder)
        slices.append((bottom, top))
        bottom = top
    return slices


@njit(cache=True)
def _update_heap(utilities: NDArray, zones: NDArray, new_u: float, new_zone: int):
    """Inserts a value into a sorted array, maintaining sort order, and the number of items in that array. The array
    is sorted lowest-to-highest"""
    i = 0
    top = len(utilities)
    while i < top:
        current_u = utilities[i]
        if new_u < current_u:
            break
        i += 1
    if i <= 0:
        return
    for j in range(i - 1):
        utilities[j] = utilities[j + 1]
        zones[j] = zones[j + 1]
    utilities[i - 1] = new_u
    zones[i - 1] = new_zone


@njit(nogil=True, cache=True)
def _nbf_twopart_worker(
    pk_costs: NDArray,
    kq_costs: NDArray,
    add_pkq_costs: NDArray,
    result_costs: NDArray,
    result_indices: NDArray,
    flag_array: NDArray,
    start: int,
    stop: int,
    n: int,
):
    """Performance-tuned Numba function to operate on its own thread"""
    n_origins, n_intermediate = pk_costs.shape
    n_destinations = kq_costs.shape[1]

    # Allocate the sorted heap of costs (and associated indices) once per thread
    cost_heap = np.full(shape=n, fill_value=_NEG_INF)
    zones_heap = np.full(shape=n, fill_value=_NULL_INDEX)

    for offset in range(start, stop):
        if not flag_array[offset]:
            continue  # Skip certain ODs
        p, q = divmod(offset, n_destinations)

        # Reset the heap for this OD
        cost_heap[:] = _NEG_INF
        zones_heap[:] = _NULL_INDEX

        for k in range(n_intermediate):
            if k >= n_intermediate:
                print("ERR", offset, p, q, k)
                raise AssertionError()

            interim_cost = pk_costs[p, k] + kq_costs[k, q]
            if add_pkq_costs is not None:
                interim_cost = interim_cost + add_pkq_costs[p, k, q]

            # In general, for problems where (n_origins * n_destinations) >> k, most values will not be in the top
            # k. So quickly check against the lowest cost in the heap to avoid calling the updater func
            if (interim_cost < cost_heap[0]) or (interim_cost == _NEG_INF):
                continue
            _update_heap(cost_heap, zones_heap, interim_cost, k)

        result_costs[p, q, :] = cost_heap
        result_indices[p, q, :] = zones_heap


@njit(nogil=True, cache=True)
def _nbf_twopart_subset_worker(
    pk_costs: NDArray,
    kq_costs: NDArray,
    add_pkq_costs: NDArray,
    subset_pk_costs: NDArray,
    result_costs: NDArray,
    result_indices: NDArray,
    flag_array: NDArray,
    start: int,
    stop: int,
    n_subset: int,
    n_final: int,
):
    """Performance-tuned Numba function to operate on its own thread"""
    n_origins, n_intermediate = pk_costs.shape
    n_destinations = kq_costs.shape[1]

    # Allocate the sorted heap of utilities (and associated indices) once per thread
    subset_cost_heap = np.full(shape=n_subset, fill_value=_NEG_INF)
    subset_zones_heap = np.full(shape=n_subset, fill_value=_NULL_INDEX)

    cost_heap = np.full(shape=n_final, fill_value=_NEG_INF)
    zone_heap = np.full(shape=n_final, fill_value=_NULL_INDEX)

    for offset in range(start, stop):
        if not flag_array[offset]:
            continue  # Skip certain ODs
        p, q = divmod(offset, n_destinations)

        # Reset the heap for this OD
        subset_cost_heap[:] = _NEG_INF
        subset_zones_heap[:] = _NULL_INDEX

        cost_heap[:] = _NEG_INF
        zone_heap[:] = _NULL_INDEX

        # Find the superset of zones by subset cost
        for k in range(n_intermediate):
            if k >= n_intermediate:
                print("ERR", offset, p, q, k)
                raise AssertionError()

            interim_pk_cost = subset_pk_costs[p, k]
            interim_cost = kq_costs[k, q]
            # This zone is not available if value is greater than zero (Given cost value should be negative now)
            if interim_pk_cost >= 0:
                continue

            # In general, for problems where (n_origins * n_destinations) >> k, most values will not be in the top
            # k. So quickly check against the lowest utility in the heap to avoid calling the updater func
            # Interim utility needs to be checked to ensure the path is a viable choice
            if (interim_pk_cost < subset_cost_heap[0]) or (interim_cost == _NEG_INF):
                continue
            _update_heap(subset_cost_heap, subset_zones_heap, interim_pk_cost, k)

        # Compute the composite utility for the subset zones
        for k in subset_zones_heap:
            interim_cost = pk_costs[p, k] + kq_costs[k, q]
            if add_pkq_costs is not None:
                interim_cost = interim_cost + add_pkq_costs[p, k, q]

            # In general, for problems where (n_origins * n_destinations) >> k, most values will not be in the top
            # k. So quickly check against the lowest utility in the heap to avoid calling the updater func
            if (interim_cost < cost_heap[0]) or (interim_cost == _NEG_INF):
                continue
            _update_heap(cost_heap, zone_heap, interim_cost, k)

        result_costs[p, q, :] = cost_heap
        result_indices[p, q, :] = zone_heap


def _validate_pk_kq_tables(pk_table: pd.DataFrame, kq_table: pd.DataFrame) -> Tuple[pd.Index, pd.Index, pd.Index]:
    if pk_table.index.nlevels != 2:
        raise RuntimeError("pk table index must have two levels")
    if kq_table.index.nlevels != 2:
        raise RuntimeError("kq table index must have two levels")

    # Take the unique index of each level, as Pandas can return more items than is present, if the frame is a slice
    origin_zones = pk_table.index.unique(level=0)
    intermediate_zones = pk_table.index.unique(level=1)
    destination_zones = kq_table.index.unique(level=1)

    # Check that the access and egress tables have compatible indices
    if not intermediate_zones.equals(kq_table.index.unique(level=0)):
        raise RuntimeError("pk index level 2 and kq index level 1 must be the same")

    return origin_zones, intermediate_zones, destination_zones


def best_intermediate_zones(
    pk_table: pd.DataFrame,
    kq_table: pd.DataFrame,
    cost_col: str,
    *,
    n: int = 1,
    add_pkq_cost: NDArray = None,
    flag_array: NDArray = None,
    maximize: bool = True,
    null_index: int = 0,
    other_columns: bool = True,
    intermediate_name: str = "intermediate_zone",
    availability_name: str = "available",
    n_threads: int = 1,
    squeeze: bool = True,
) -> Union[pd.DataFrame, Dict[int, pd.DataFrame]]:
    """Numba-accelerated.

    Triple-index operation for two matrices, finding the most- or least-cost intermediate zones. Takes a first leg
    matrix ("pk") and a second leg matrix ("kq") to produce a combined "pq" matrix with the best intermediate "k".
    Also works to construct multiple "pq" matrices for the top _n_ intermediate "k" zones.

    There is no restriction on the label dtypes, as long as the "pk" (leg 1) and "kq" (leg 2) tables share the same
    "k" index.

    Both the input matrices must be provided in "tall" format - as Pandas Series with a 2-level MultiIndex.
    Essentially, the "pk" (leg 1) and "kq" (leg 2) tables are DataFrames with multiple matrices defined within. The
    output table(s) are also returned in a tall format.

    When constructing the result tables, columns in the "pk" (leg 1) and "kq" (leg 2) tables are "carried forward"
    such that the results columns will be the union of columns in the input tables. Columns in one table only will
    be carried forward unmodified and retain their data type. Columns in both tables will be added together, and
    thus MUST be numeric.

    In the specified cost column, a value of `-inf` (or `inf` when minimizing) is respected as the sentinel value
    for unavailable. "pk" or "kq" interchanges with this sentinel value will not be considered.

    Args:
        pk_table (pd.DataFrame): A DataFrame with 2-level MultiIndex of the shape ((p, k), A). Must include the
            specified cost column
        kq_table (pd.DataFrame): A DataFrame with 2-level MultiIndex of the shape ((k, q), E). Must include the
            specified cost column
        cost_col (str): Name of the column in the access and egress table to use as the cost to
            minimize/maximize. Values of `+/- inf` are respected to indicate unavailable choices.
        n (int, optional): Defaults to ``1``. The number of ranks to return (e.g., find the _n_ best intermediate
            zones). If n <= 0, it will be corrected to 1.
        add_pkq_cost (ndarray): A 3-dimensional numpy array containing additional two-part path ("pkq") costs to be
            included in triple-index operation.
        flag_array (ndarray, optional): Defaults to ``None``. An array of boolean flags indicating the "pq"
            zone pairs to evaluate.
        maximize (bool, optional): Defaults to ``True``. If True, this function maximize the result. If False, it
            minimizes it.
        null_index (int, optional): Defaults to ``0``. Fill value used if NO intermediate zone is available.
        other_columns (bool, optional): Defaults to ``True``. If True, the result DataFrame will include all columns
            in the "pk" and "kq" tables. The result table will be of the shape ((p, q), A | E + 3)
        intermediate_name (str, optional): Defaults to ``'intermediate_zone'``. Name of the column in the result
            table containing the selected intermediate zone.
        availability_name (str, optional): Defaults to ``'available'``. Name of the column in the result table
            containing a flag whether ANY intermediate zone was found to be available.
        n_threads (int, optional): Defaults to ``1``. Number of threads to use.
        squeeze (bool, optional): Defaults to ``True``. If ``n == 1`` and ``squeeze=True``, a single DataFrame is
            returned. Otherwise, a dictionary of DataFrames will be returned.

    Returns:
        DataFrame: If n == 1 and squeeze=True. A DataFrame of the shape ((p, q), A | E + 3), containing the
            intermediate "k" zone selected, the associated max/min cost, and a flag indicating its availability.
            Additional columns from the "pk" (leg 1) and "kq" (leg 2) tables, indexed for the appropriately chosen
            intermediate zone, will also be included if other_columns=True.
        Dict[int, DataFrame]: If n > 1. The keys represent the ranks, so result[1] is the best intermediate zone,
            result[2] is the second-best, etc. The value DataFrames are in the same format as if n == 1, just with
            different intermediate zones chosen.
    """

    # Check inputs
    n = max(1, n)
    n_threads = int(np.clip(n_threads, 1, cpu_count()))

    origin_zones, intermediate_zones, destination_zones = _validate_pk_kq_tables(pk_table, kq_table)

    # Set up the raw data
    n_origins, n_intermediate, n_destinations = len(origin_zones), len(intermediate_zones), len(destination_zones)
    pk_cost = pk_table[cost_col].to_numpy().reshape([n_origins, n_intermediate]).astype(np.float64)
    kq_cost = kq_table[cost_col].to_numpy().reshape([n_intermediate, n_destinations]).astype(np.float64)
    if not maximize:  # Invert the cost to use code set to maximize
        pk_cost *= -1
        kq_cost *= -1

    result_cost = np.zeros(shape=(n_origins, n_destinations, n), dtype=np.float64)
    result_indices = np.zeros(shape=(n_origins, n_destinations, n), dtype=np.int16)

    if add_pkq_cost is not None:
        if add_pkq_cost.shape != (n_origins, n_intermediate, n_destinations):
            raise RuntimeError(
                "Shape of `add_pkq_cost` does not match the number of origin, intermediate, and "
                "destination zones found in `pk_table` and `kq_table`"
            )
        add_pkq_cost = add_pkq_cost.astype(np.float64)

    if flag_array is None:
        flag_array = np.full(n_origins * n_destinations, fill_value=True)
    else:
        if not len(flag_array) == (n_origins * n_destinations):
            raise RuntimeError("Length of `flag_array` incompatible with size of `pk_table` and `kq_table`")
        flag_array = flag_array.astype(np.bool_)

    # Run the accelerated function
    breaks = _get_breaks(n_origins * n_destinations, n_threads)
    threads = np.empty(n_threads, dtype="O")
    for i, (start, stop) in enumerate(breaks):
        t = Thread(
            target=_nbf_twopart_worker,
            args=[pk_cost, kq_cost, add_pkq_cost, result_cost, result_indices, flag_array, start, stop, n],
        )
        t.start()
        threads[i] = t
    for t in threads:
        t.join()

    return _combine_tables(
        pk_table,
        kq_table,
        result_cost,
        result_indices,
        flag_array,
        origin_zones,
        intermediate_zones,
        destination_zones,
        cost_col,
        availability_name,
        intermediate_name,
        other_columns=other_columns,
        squeeze=squeeze,
        null_index=null_index,
    )


def best_intermediate_subset_zones(
    pk_subset_cost: pd.Series,
    pk_table: pd.DataFrame,
    kq_table: pd.DataFrame,
    cost_col: str,
    *,
    n_subset: int = 1,
    n_final: int = 1,
    add_pkq_cost: NDArray = None,
    flag_array: NDArray = None,
    maximize_subset: bool = True,
    maximize_final: bool = True,
    null_index: int = 0,
    other_columns: bool = True,
    intermediate_name: str = "intermediate_zone",
    availability_name: str = "available",
    n_threads: int = 1,
    squeeze: bool = True,
) -> Union[pd.DataFrame, Dict[int, pd.DataFrame]]:
    """Numba-accelerated.

    Triple-index operation for two matrices, finding the most- or least-cost intermediate zones from a subset.
    Takes a first leg matrix ("pk") and a second leg matrix ("kq") to produce a combined "pq" matrix with the best
    intermediate "k". Also works to construct multiple "pq" matrices for the top _n_final_ intermediate "k" zones.

    There is no restriction on the label dtypes, as long as the "pk" (leg 1) and "kq" (leg 2) tables share the same
    "k" index.

    Both the input matrices must be provided in "tall" format - as Pandas Series with a 2-level MultiIndex.
    Essentially, the "pk" (leg 1) and "kq" (leg 2) tables are DataFrames with multiple matrices defined within. The
    output table(s) are also returned in a tall format.

    When constructing the result tables, columns in the "pk" (leg 1) and "kq" (leg 2) tables are "carried forward"
    such that the results columns will be the union of columns in the input tables. Columns in one table only will
    be carried forward unmodified and retain their data type. Columns in both tables will be added together, and
    thus MUST be numeric.

    In the specified cost column, a value of `-inf` (or `inf` when minimizing) is respected as the sentinel value
    for unavailable. "pk" or "kq" interchanges with this sentinel value will not be considered.

    Args:
        pk_subset_cost (pd.Series): A Series with 2-level MultiIndex containing values to use for subsetting
        pk_table (pd.DataFrame): A DataFrame with 2-level MultiIndex of the shape ((p, k), A). Must include the
            specified cost column
        kq_table (pd.DataFrame): A DataFrame with 2-level MultiIndex of the shape ((k, q), E). Must include the
            specified cost column
        cost_col (str): Name of the column in the access and egress table to use as the cost to
            minimize/maximize. Values of `+/- inf` are respected to indicate unavailable choices.
        n_subset (int, optional): Defaults to ``1``. The number of intermediate ranks to subset (e.g., find the
            _n_subset_ best intermediate zones to select the _n_final_ best intermediate zones). If n_subset <= 0,
            it will be corrected to 1.
        n_final (int, optional): Defaults to ``1``. The number of ranks to return (e.g., find the _n_final_ best
            intermediate zones). If n_final <= 0, it will be corrected to 1.
        add_pkq_cost (ndarray): A 3-dimensional numpy array containing additional two-part path ("pkq") costs to be
            included in triple-index operation.
        flag_array (ndarray, optional): Defaults to ``None``. An array of boolean flags indicating the "pq"
            zone pairs to evaluate.
        maximize_subset (bool, optional): Defaults to ``True``. If True, this function maximize the result when
            determining the "pk" selection. If False, it minimizes it.
        maximize_final (bool, optional): Defaults to ``True``. If True, this function maximize the result when
            determining the "pq" selection. If False, it minimizes it.
        null_index (int, optional): Defaults to ``0``. Fill value used if NO intermediate zone is available.
        other_columns (bool, optional): Defaults to ``True``. If True, the result DataFrame will include all columns
            in the "pk" and "kq" tables. The result table will be of the shape ((p, q), A | E + 3)
        intermediate_name (str, optional): Defaults to ``'intermediate_zone'``. Name of the column in the result
            table containing the selected intermediate zone.
        availability_name (str, optional): Defaults to ``'available'``. Name of the column in the result table
            containing a flag whether ANY intermediate zone was found to be available.
        n_threads (int, optional): Defaults to ``1``. Number of threads to use.
        squeeze (bool, optional): Defaults to ``True``. If ``n_final == 1`` and ``squeeze=True``, a single DataFrame
            is returned. Otherwise, a dictionary of DataFrames will be returned.

    Returns:
        DataFrame: If n_final == 1 and squeeze=True. A DataFrame of the shape ((p, q), A | E + 3), containing the
            intermediate "k" zone selected, the associated max/min cost, and a flag indicating its availability.
            Additional columns from the "pk" (leg 1) and "kq" (leg 2) tables, indexed for the appropriately chosen
            intermediate zone, will also be included if other_columns=True.
        Dict[int, DataFrame]: If n_final > 1. The keys represent the ranks, so result[1] is the best intermediate
            zone, result[2] is the second-best, etc. The value DataFrames are in the same format as if n_final == 1,
            just with different intermediate zones chosen.
    """

    # Check inputs
    n_subset = max(1, n_subset)
    n_final = max(1, n_final)
    if n_subset < n_final:
        raise ValueError("`n_subset` must be >= `n_final`")
    n_threads = int(np.clip(n_threads, 1, cpu_count()))

    origin_zones, intermediate_zones, destination_zones = _validate_pk_kq_tables(pk_table, kq_table)
    if not pk_subset_cost.index.equals(pk_table.index):
        raise RuntimeError("`pk_subset_cost` and `pk_table` have incompatible indices")

    # Set up the raw data
    n_origins, n_intermediate, n_destinations = len(origin_zones), len(intermediate_zones), len(destination_zones)
    pk_subset_cost = pk_subset_cost.to_numpy().reshape([n_origins, n_intermediate]).astype(np.float64)
    if not maximize_subset:  # Invert the cost to use code set to maximize
        pk_subset_cost *= -1
    pk_cost = pk_table[cost_col].to_numpy().reshape([n_origins, n_intermediate]).astype(np.float64)
    kq_cost = kq_table[cost_col].to_numpy().reshape([n_intermediate, n_destinations]).astype(np.float64)
    if not maximize_final:  # Invert the cost to use code set to maximize
        pk_cost *= -1
        kq_cost *= -1

    result_cost = np.zeros(shape=(n_origins, n_destinations, n_final), dtype=np.float64)
    result_indices = np.zeros(shape=(n_origins, n_destinations, n_final), dtype=np.int16)

    if add_pkq_cost is not None:
        if add_pkq_cost.shape != (n_origins, n_intermediate, n_destinations):
            raise RuntimeError(
                "Shape of `add_pkq_cost` does not match the number of origin, intermediate, and "
                "destination zones found in `pk_table` and `kq_table`"
            )
        add_pkq_cost = add_pkq_cost.astype(np.float64)

    if flag_array is None:
        flag_array = np.full(n_origins * n_destinations, fill_value=True)
    else:
        if not len(flag_array) == (n_origins * n_destinations):
            raise RuntimeError("Length of `flag_array` incompatible with size of `pk_table` and `kq_table`")
        flag_array = flag_array.astype(np.bool_)

    # Run the accelerated function
    breaks = _get_breaks(n_origins * n_destinations, n_threads)
    threads = np.empty(n_threads, dtype="O")
    for i, (start, stop) in enumerate(breaks):
        t = Thread(
            target=_nbf_twopart_subset_worker,
            args=[
                pk_cost,
                kq_cost,
                add_pkq_cost,
                pk_subset_cost,
                result_cost,
                result_indices,
                flag_array,
                start,
                stop,
                n_subset,
                n_final,
            ],
        )
        t.start()
        threads[i] = t
    for t in threads:
        t.join()

    return _combine_tables(
        pk_table,
        kq_table,
        result_cost,
        result_indices,
        flag_array,
        origin_zones,
        intermediate_zones,
        destination_zones,
        cost_col,
        availability_name,
        intermediate_name,
        other_columns=other_columns,
        squeeze=squeeze,
        null_index=null_index,
    )


def _reshape_series(
    series: pd.Series, n_rows: int, n_cols: int, row_indexer: NDArray, col_indexer: NDArray
) -> Union[NDArray, pd.Categorical]:
    if isinstance(series, pd.CategoricalDtype):
        codes = series.cat.codes
        reshaped = codes.to_numpy().reshape([n_rows, n_cols])
        flat = reshaped[row_indexer, col_indexer]
        return pd.Categorical.from_codes(flat, categories=series.cat.categories)
    else:
        reshaped = series.to_numpy().reshape([n_rows, n_cols])
        return reshaped[row_indexer, col_indexer]


def _combine_tables(
    pk_table: pd.DataFrame,
    kq_table: pd.DataFrame,
    result_cost: NDArray,
    result_indices: NDArray,
    flag_array: NDArray,
    origin_zones: pd.Index,
    intermediate_zones: pd.Index,
    destination_zones: pd.Index,
    cost_col: str,
    avail_col: str,
    intermediate_name: str,
    *,
    other_columns: bool = True,
    squeeze: bool = True,
    null_index: int = 0,
) -> Union[pd.DataFrame, Dict[int, pd.DataFrame]]:
    n_origins, n_intermediate, n_destinations = len(origin_zones), len(intermediate_zones), len(destination_zones)
    n_selected: int = result_indices.shape[2]

    # Construct composite result tables
    if other_columns:
        remaining_columns = set(pk_table.columns.union(kq_table.columns)) - {cost_col, intermediate_name, avail_col}
    else:
        remaining_columns = set()

    # Collect columns for the number of selected zones
    row_index = pd.MultiIndex.from_product([origin_zones, destination_zones])
    pk_indexer = np.repeat(np.arange(n_origins), n_destinations)
    kq_indexer = np.tile(np.arange(n_destinations), n_origins)
    if flag_array is not None:
        row_index = row_index[flag_array]
        pk_indexer = pk_indexer[flag_array]
        kq_indexer = kq_indexer[flag_array]

    tables = {}
    for i in range(n_selected):
        table = pd.DataFrame(index=row_index)

        offsets_i = result_indices[:, :, i]  # 2D indexer for this (i ∈ k) path
        flat_offsets_i = offsets_i.flatten()  # Convert to 1D indexer
        raw_cost = result_cost[:, :, i].flatten()
        if flag_array is not None:
            flat_offsets_i = flat_offsets_i[flag_array]
            raw_cost = raw_cost[flag_array]

        availability_i = flat_offsets_i != _NULL_INDEX
        intermediate_result_i = intermediate_zones.to_numpy().take(flat_offsets_i)
        intermediate_result_i[~availability_i] = null_index

        table[intermediate_name] = intermediate_result_i
        table[avail_col] = availability_i
        table[cost_col] = raw_cost

        # Loop through any additional columns, adding together if they exist in both tables
        for column in sorted(remaining_columns):
            in_pk_table = column in pk_table
            in_kq_table = column in kq_table

            pk_component = 0
            if in_pk_table:
                pk_component = _reshape_series(pk_table[column], n_origins, n_intermediate, pk_indexer, flat_offsets_i)

            kq_component = 0
            if in_kq_table:
                kq_component = _reshape_series(
                    kq_table[column], n_intermediate, n_destinations, flat_offsets_i, kq_indexer
                )

            if in_pk_table and in_kq_table:
                composite_data = pk_component + kq_component
            elif in_pk_table:
                composite_data = pk_component
            elif in_kq_table:
                composite_data = kq_component
            else:
                raise RuntimeError("This shouldn't happen")
            table[column] = composite_data

        tables[n_selected - i] = table

    if (n_selected == 1) and squeeze:
        return tables[1]
    return tables



import unittest

import numpy as np
import pandas as pd
from pandas import testing as pdt

from wsp_balsa.routines import (aggregate_matrix, matrix_balancing_1d,
                                matrix_balancing_2d, matrix_bucket_rounding)


class TestMatrixBucketRounding(unittest.TestCase):

    def test_small(self):
        """ Test bucket rounding routine on a small matrix to various levels of rounding precision. """
        a = np.random.uniform(0, 1000, (5, 5))
        for decimal in range(-2, 6):
            a_rnd = matrix_bucket_rounding(a, decimals=decimal)

            self._compare_matrix_sums(a_rnd, a, decimal)
            self._compare_matrix_values(a_rnd, a, decimal)

    def test_return_type(self):
        """ Test that bucket rounding returns an integer or float dtype, where appropriate. """
        a = np.random.uniform(0, 1000, (5, 5))

        # first test, float return
        b = matrix_bucket_rounding(a, decimals=2)
        self.assertEqual(b.dtype, a.dtype, "dtype of bucket rounded matrix is not equal to dtype of input matrix")

        # second test, int return
        b = matrix_bucket_rounding(a, decimals=0)
        self.assertEqual(b.dtype, np.dtype('int32'), "dtype of bucket rounded matrix is not integer")

    def test_large(self):
        """ Test bucket rounding routine on a large matrix to various levels of rounding precision. """
        a = np.random.uniform(0, 1000, (1000, 1000))
        for decimal in [-2, 0, 5]:
            a_rnd = matrix_bucket_rounding(a, decimals=decimal)

            self._compare_matrix_sums(a_rnd, a, decimal)
            self._compare_matrix_values(a_rnd, a, decimal)

    def test_pandas_import(self):
        """Test return type and values if matrix is passed as a Pandas DataFrame. """
        a = np.random.uniform(0, 1000, (5, 5))
        df = pd.DataFrame(a)
        decimals = 3
        df_rnd = matrix_bucket_rounding(df, decimals=decimals)

        self._compare_matrix_sums(df.values, df_rnd.values, decimals)
        self._compare_matrix_values(df.values, df_rnd.values, decimals)
        self.assertEqual(type(df_rnd), pd.DataFrame, "dtype of returned matrix is a Pandas DataFrame")

    def _compare_matrix_sums(self, a, b, decimal):
        max_error = 0.5 * (10.0 ** (-decimal))
        a_sum = np.sum(a)
        b_sum = np.sum(b)

        self.assertLessEqual(a_sum, b_sum + max_error, "Bucket rounded matrix is not within a small margin of error")
        self.assertGreaterEqual(a_sum, b_sum - max_error, "Bucket rounded matrix is not within a small margin of error")

    def _compare_matrix_values(self, a, b, decimal):
        max_error = 10.0 ** (-decimal)
        np.testing.assert_allclose(a, b, atol=max_error, rtol=0.0,
                                   err_msg="Bucket rounded matrix values are not within %f" % (max_error))


class TestAggregateMatrix(unittest.TestCase):

    def setUp(self):
        self._square_symatrix = np.array([
            [4, 9, 1, 1, 9, 0, 4, 4, 9],
            [10, 0, 0, 10, 2, 8, 0, 10, 9],
            [8, 9, 9, 6, 7, 7, 4, 1, 8],
            [9, 8, 1, 10, 6, 7, 2, 1, 2],
            [3, 3, 10, 5, 3, 9, 7, 9, 4],
            [1, 5, 1, 1, 7, 4, 2, 9, 0],
            [4, 3, 4, 1, 5, 9, 3, 7, 5],
            [2, 3, 7, 2, 2, 10, 2, 3, 5],
            [5, 10, 4, 9, 1, 5, 4, 4, 7]
        ])
        self._numeric_index = pd.Index([10, 11, 12, 20, 21, 22, 30, 31, 32])
        self._square_symatrix = pd.DataFrame(self._square_symatrix, index=self._numeric_index, columns=self._numeric_index)
        self._tall_symatrix = self._square_symatrix.stack()

        self._square_dmatrix = np.array([
            [1, 3, 2, 9],
            [5, 0, 4, 7],
            [10, 5, 4, 0],
            [6, 1, 8, 9],
            [10, 5, 0, 3],
            [4, 8, 8, 6],
            [1, 0, 4, 1],
            [2, 3, 4, 6],
            [3, 0, 0, 7],
        ])
        self._text_index = pd.Index(['A1', 'A2', 'B1', 'B2'])
        self._square_dmatrix = pd.DataFrame(self._square_dmatrix, index=self._numeric_index, columns=self._text_index)
        self._tall_dmatrix = self._square_dmatrix.stack()

        self._grouper1 = pd.Series({
            10: 1,
            11: 1,
            12: 1,
            20: 2,
            21: 2,
            22: 2,
            30: 3,
            31: 3,
            32: 3
        })

        self._grouper2 = pd.Series({
            'A1': 'A',
            'A2': 'A',
            'B1': 'B',
            'B2': 'B'
        })

    def test_square_symatrix(self):
        expected_result = pd.DataFrame([
            [50, 50, 49],
            [41, 52, 36],
            [42, 44, 40]
        ], index=[1, 2, 3], columns=[1, 2, 3])

        test1 = aggregate_matrix(self._square_symatrix, row_groups=self._grouper1, col_groups=self._grouper1)
        pdt.assert_frame_equal(expected_result, test1, check_dtype=False)

        test2 = aggregate_matrix(self._square_symatrix, groups=self._grouper1)
        pdt.assert_frame_equal(expected_result, test2, check_dtype=False)

        test3 = aggregate_matrix(self._square_symatrix, groups=self._grouper1.values)
        pdt.assert_frame_equal(expected_result, test3, check_dtype=False)

    def test_square_dmatrix(self):
        expected_result = pd.DataFrame([
            [24, 26],
            [34, 34],
            [9, 22],
        ], index=[1, 2, 3], columns=['A', 'B'])

        test1 = aggregate_matrix(self._square_dmatrix, row_groups=self._grouper1, col_groups=self._grouper2)
        pdt.assert_frame_equal(expected_result, test1, check_dtype=False)

        test2 = aggregate_matrix(self._square_dmatrix, row_groups=self._grouper1.values,
                                 col_groups=self._grouper2.values)
        pdt.assert_frame_equal(expected_result, test2, check_dtype=False)

    def test_tall_symatrix(self):
        expected_result = pd.DataFrame([
            [1, 1, 50],
            [1, 2, 50],
            [1, 3, 49],
            [2, 1, 41],
            [2, 2, 52],
            [2, 3, 36],
            [3, 1, 42],
            [3, 2, 44],
            [3, 3, 40],
        ], columns=['row', 'col', 'val']).set_index(['row', 'col'])['val']

        tall_row_grouper = self._grouper1.reindex(self._tall_symatrix.index, level=0)
        tall_col_grouper = self._grouper1.reindex(self._tall_symatrix.index, level=1)

        test1 = aggregate_matrix(self._tall_symatrix, row_groups=tall_row_grouper, col_groups=tall_col_grouper)
        pdt.assert_series_equal(expected_result, test1, check_dtype=False, check_names=False)

        test2 = aggregate_matrix(self._tall_symatrix, row_groups=self._grouper1, col_groups=self._grouper1)
        pdt.assert_series_equal(expected_result, test2, check_dtype=False, check_names=False)

        test3 = aggregate_matrix(self._tall_symatrix,
                                 row_groups=tall_row_grouper.values, col_groups=tall_col_grouper.values)
        pdt.assert_series_equal(expected_result, test3, check_dtype=False, check_names=False)

        test4 = aggregate_matrix(self._tall_symatrix, groups=self._grouper1)
        pdt.assert_series_equal(expected_result, test4, check_dtype=False, check_names=False)

    def test_tall_dmatrix(self):
        expected_result = pd.DataFrame([
            [1, 'A', 24],
            [1, 'B', 26],
            [2, 'A', 34],
            [2, 'B', 34],
            [3, 'A', 9],
            [3, 'B', 22]
        ], columns=['row', 'col', 'val']).set_index(['row', 'col'])['val']

        tall_row_grouper = self._grouper1.reindex(self._tall_dmatrix.index, level=0)
        tall_col_grouper = self._grouper2.reindex(self._tall_dmatrix.index, level=1)

        test1 = aggregate_matrix(self._tall_dmatrix, row_groups=tall_row_grouper, col_groups=tall_col_grouper)
        pdt.assert_series_equal(expected_result, test1, check_dtype=False, check_names=False)

        test2 = aggregate_matrix(self._tall_dmatrix, row_groups=self._grouper1, col_groups=self._grouper2)
        pdt.assert_series_equal(expected_result, test2, check_dtype=False, check_names=False)

        test3 = aggregate_matrix(self._tall_dmatrix,
                                 row_groups=tall_row_grouper.values, col_groups=tall_col_grouper.values)
        pdt.assert_series_equal(expected_result, test3, check_dtype=False, check_names=False)


class TestMatrixBalancing(unittest.TestCase):
    def setUp(self):
        self._square_matrix = np.random.uniform(0, 1000, (5, 5))
        self._1darray = np.random.uniform(0, 1000, 5)

    def test_1d_balance(self):
        axes = [0, 1]
        for ax in axes:
            test = matrix_balancing_1d(self._square_matrix, self._1darray, ax)
            self.assertAlmostEqual(test.sum(), self._1darray.sum(), places=5)
            pdt.assert_series_equal(pd.Series(np.sum(test, ax)), pd.Series(self._1darray))

    def test_2d_balance_matched_total(self):
        row = self._1darray
        column = np.roll(self._1darray, 2)

        test = matrix_balancing_2d(self._square_matrix, row, column, rel_error=0.000001)
        pdt.assert_series_equal(pd.Series(np.sum(test[0], 1)), pd.Series(row), check_exact=False)
        pdt.assert_series_equal(pd.Series(np.sum(test[0], 0)), pd.Series(column), check_exact=False)

    def test_2d_balance_average_total(self):
        row = self._1darray
        column = np.roll(np.sqrt(row), 2)

        test = matrix_balancing_2d(self._square_matrix, row, column, rel_error=0.000001, totals_to_use='average')
        self.assertAlmostEqual(test[0].sum().sum(), (row.sum() + column.sum())/2, places=5)

    def test_2d_balance_row_total(self):
        row = self._1darray
        column = np.sqrt(row)

        test = matrix_balancing_2d(self._square_matrix, row, column, rel_error=0.000001, totals_to_use='rows')
        pdt.assert_series_equal(pd.Series(np.sum(test[0], 1)), pd.Series(row), check_exact=False)

    def test_2d_balance_col_total(self):
        row = self._1darray
        column = np.sqrt(row)

        test = matrix_balancing_2d(self._square_matrix, row, column, rel_error=0.000001, totals_to_use='columns')
        pdt.assert_series_equal(pd.Series(np.sum(test[0], 0)), pd.Series(column), check_exact=False)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

__all__ = [
    "read_omx",
    "to_omx",
]

from os import PathLike
from typing import Dict, Iterable, List, Tuple, Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from ..general import sort_nicely
from ..matrices import fast_unstack

try:
    import openmatrix as omx
except ImportError:
    omx = None

MATRIX_TYPES = Union[pd.DataFrame, pd.Series, NDArray]


if omx is None:

    def read_omx(*args, **kwargs):
        raise NotImplementedError()

    def to_omx(*args, **kwargs):
        raise NotImplementedError()

else:

    def read_omx(
        src_fp: Union[str, PathLike],
        *,
        tables: Iterable[str] = None,
        mapping: str = None,
        tall: bool = False,
        raw: bool = False,
        squeeze: bool = True,
    ) -> Union[MATRIX_TYPES, Dict[str, MATRIX_TYPES]]:
        """
        Reads Open Matrix (OMX) files. An OMX file can contain multiple matrices, so this function
        typically returns a Dict.

        Args:
            src_fp (str | PathLike): OMX file from which to read. Cannot be an open file handler.
            tables (Iterable[str], optional): List of matrices to read from the file. If None, all matrices will be
                read.
            mapping (str, optional): The zone number mapping to use, if known in advance.
            tall (bool, optional) : If True, matrices will be returned in 1D format. Otherwise, a 2D object is returned.
            raw (bool, optional): If True, matrices will be returned as raw Numpy arrays. Otherwise, Pandas objects are
                returned
            squeeze (bool, optional): If True, and the file contains exactly one matrix, return that matrix instead of a
                Dict.

        Returns:
            The matrix, or matrices contained in the OMX file.

        """
        omx_file = omx.open_file(str(src_fp), mode="r")
        try:
            table_names: List[str] = sort_nicely(omx_file.list_matrices()) if tables is None else list(tables)

            if not raw:
                if mapping is None:
                    rows, columns = omx_file.shape()
                    if rows != columns:
                        raise NotImplementedError("Handling of non-square matrices not implemented yet")
                    labels = pd.Index(range(rows))
                else:
                    zone_mapping: Dict[int, int] = omx_file.mapping(mapping)
                    labels = pd.Index(zone_mapping.keys())
                if tall:
                    labels = pd.MultiIndex.from_product([labels, labels], names=["o", "d"])

            retval = {}
            for name in table_names:
                matrix = np.array(omx_file[name])
                if tall:
                    n = matrix.shape[0] * matrix.shape[1]
                    matrix.shape = n

                if not raw:
                    if tall:
                        matrix = pd.Series(matrix, index=labels, name=name)
                    else:
                        matrix = pd.DataFrame(matrix, index=labels, columns=labels)

                retval[name] = matrix

            if (len(retval) == 1) and squeeze:
                return retval[table_names[0]]
            return retval
        finally:
            omx_file.close()

    def to_omx(
        dst_fp: Union[str, PathLike],
        tables: Dict[str, MATRIX_TYPES],
        *,
        zone_index: pd.Index = None,
        title: str = "",
        descriptions: Dict[str, str] = None,
        attrs: Dict[str, Dict] = None,
        mapping_name: str = "zone_numbers",
    ):
        """Creates a new (or overwrites an old) OMX file with a collection of matrices.

        Args:
            dst_fp (str | PathLike): OMX to write.
            tables (Dict[str, pd.DataFrame | pd.Series | np.ndarray]: Collection of matrices to write. MUST be a dict,
                to permit the encoding of matrix metadata, and must contain the same types: all Numpy arrays, all
                Series, or all DataFrames. Checking is done to ensure that all items have the same shape and labels.
            zone_index: (pd.Index, optional): Override zone labels to use. Generally only useful if writing a dict of
                raw Numpy arrays.
            title (str, optional): The title saved in the OMX file.
            descriptions (Dict[str, str], optional): A dict of descriptions (one for each given matrix).
            attrs (Dict[str, Dict], optional): A dict of dicts (one for each given matrix).
            mapping_name (str, optional): Name of the mapping internal to the OMX file
        """

        matrices, zone_index = _prep_matrix_dict(tables, zone_index)

        if descriptions is None:
            descriptions = {name: "" for name in matrices.keys()}
        if attrs is None:
            attrs = {name: None for name in matrices.keys()}

        omx_file = omx.open_file(str(dst_fp), mode="w", title=title)
        try:
            omx_file.create_mapping(mapping_name, zone_index.tolist())
            for name, array in matrices.items():
                description = descriptions[name]
                attr = attrs[name]
                omx_file.create_matrix(name, obj=np.ascontiguousarray(array), title=description, attrs=attr)
        finally:
            omx_file.close()

    def _prep_matrix_dict(
        matrices: Dict[str, MATRIX_TYPES], desired_zone_index: pd.Index
    ) -> Tuple[Dict[str, np.ndarray], pd.Index]:
        collection_type = _check_types(matrices)

        if collection_type == "RAW":
            checked, n = _check_raw_matrices(matrices)
            zone_index = pd.Index(range(n))
        elif collection_type == "SERIES":
            checked, zone_index = _check_matrix_series(matrices)
        elif collection_type == "FRAME":
            checked, zone_index = _check_matrix_frames(matrices)
        else:
            raise NotImplementedError(collection_type)

        if desired_zone_index is not None:
            assert desired_zone_index.equals(zone_index)

        return checked, zone_index

    def _check_types(matrices: Dict[str, MATRIX_TYPES]) -> str:
        gen = iter(matrices.values())
        first = next(gen)

        item_type = "RAW"
        if isinstance(first, pd.Series):
            item_type = "SERIES"
        elif isinstance(first, pd.DataFrame):
            item_type = "FRAME"

        msg = "All items must be the same type"

        for item in gen:
            if item_type == "FRAME":
                assert isinstance(item, pd.DataFrame), msg
            elif item_type == "SERIES":
                assert isinstance(item, pd.Series), msg
            else:
                assert isinstance(item, np.ndarray), msg

        return item_type

    def _check_raw_matrices(matrices: Dict[str, np.ndarray]) -> Tuple[Dict[str, np.ndarray], int]:
        gen = iter(matrices.items())
        name, matrix = next(gen)

        n_dim = len(matrix.shape)
        if n_dim == 1:
            shape = matrix.shape[0]
            n = int(shape**0.5)
            assert n * n == shape, "Only tall matrices that decompose to square shapes are permitted."
            matrix = matrix[...]
            matrix.shape = n, n
        elif n_dim == 2:
            n, cols = matrix.shape
            assert n == cols, "Only square matrices are permitted"
        else:
            raise TypeError("Only 1D and 2D arrays can be saved to OMX files")

        retval = {name: matrix}
        for name, matrix in gen:
            assert len(matrix.shape) == n_dim

            if n_dim == 1:
                assert matrix.shape[0] == (n * n)
                matrix = matrix[...]
                matrix.shape = n, n
            else:
                assert matrix.shape == (n, n)

            retval[name] = matrix

        return retval, n

    def _check_matrix_series(matrices: Dict[str, pd.Series]) -> Tuple[Dict[str, np.ndarray], pd.Index]:
        gen = iter(matrices.items())
        name, matrix = next(gen)

        tall_index = matrix.index
        assert tall_index.nlevels == 2
        matrix = matrix.unstack()
        zone_index = matrix.index
        assert zone_index.equals(matrix.columns)

        retval = {name: matrix.to_numpy()}
        for name, matrix in gen:
            assert tall_index.equals(matrix.index)
            matrix = fast_unstack(matrix, zone_index, zone_index).to_numpy()
            retval[name] = matrix

        return retval, zone_index

    def _check_matrix_frames(matrices: Dict[str, pd.DataFrame]) -> Tuple[Dict[str, np.ndarray], pd.Index]:
        gen = iter(matrices.items())
        name, matrix = next(gen)

        zone_index = matrix.index
        assert zone_index.equals(matrix.columns)

        retval = {name: matrix.to_numpy()}
        for name, matrix in gen:
            assert zone_index.equals(matrix.index)
            assert zone_index.equals(matrix.columns)
            retval[name] = matrix.to_numpy()

        return retval, zone_index



from __future__ import annotations

__all__ = [
    "coerce_matrix",
    "expand_array",
    "open_file",
]

from contextlib import contextmanager
from io import FileIO
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray


def coerce_matrix(
    matrix: Union[NDArray, pd.DataFrame, pd.Series],
    *,
    allow_raw: bool = True,
    force_square: bool = True,
) -> NDArray:
    """Infers a NumPy array from given input

    Args:
        matrix (NDArray | pandas.DataFrame | pandas.Series): Matrix data to coerce its shape into a 2D ndarray
        allow_raw (bool, optional): Defaults to ``True``.
        force_square (bool, optional): Defaults to ``True``. A flag to check if the resulting matrix is square in shape

    Returns:
        NDArray: A 2D ndarray of type float32
    """
    if isinstance(matrix, pd.DataFrame):
        if force_square:
            if not matrix.index.equals(matrix.columns):
                raise ValueError("Rows and columns in `matrix` are not identical")
        return matrix.values.astype(np.float32)
    elif isinstance(matrix, pd.Series):
        if matrix.index.nlevels != 2:
            raise ValueError("Cannot infer a matrix from a Series with more or fewer than 2 levels")
        wide = matrix.unstack()

        union = wide.index | wide.columns
        wide = wide.reindex_axis(union, fill_value=0.0, axis=0).reindex_axis(union, fill_value=0.0, axis=1)
        return wide.values.astype(np.float32)

    if not allow_raw:
        raise NotImplementedError()

    matrix = np.array(matrix, dtype=np.float32)
    if len(matrix.shape) != 2:
        raise ValueError("`matrix` is not 2-dimensional")
    i, j = matrix.shape
    if i != j:
        raise ValueError("`matrix` is not square")

    return matrix


def expand_array(
    a: NDArray,
    n: NDArray,
    *,
    axis: int = None,
) -> NDArray:
    """Expands an array across all dimensions by a set amount

    Args:
        a (NDArray): The array to expand
        n (NDArray): The (non-negative) number of items to expand by.
        axis (int, optional): Defaults to ``None``. The axis to expand along, or None to expand along all axes.

    Returns:
        NDArray: The expanded array
    """

    if axis is None:
        new_shape = [dim + n for dim in a.shape]
    else:
        new_shape = []
        for i, dim in enumerate(a.shape):
            dim += n if i == axis else 0
            new_shape.append(dim)

    out = np.zeros(new_shape, dtype=a.dtype)

    indexer = [slice(0, dim) for dim in a.shape]
    out[indexer] = a

    return out


@contextmanager
def open_file(file_handle: Union[str, Path, FileIO], **kwargs):
    """Context manager for opening files provided as several different types. Supports a file handler as a str, unicode,
    ``pathlib.Path``, or an already-opened handler.

    Args:
        file_handle (str | Path | FileIO): The item to be opened or is already open.
        **kwargs: Keyword args passed to ``open()``. Usually mode='w'.

    Yields:
        File:
            The opened file handler. Automatically closed once out of context.
    """
    opened = False
    if isinstance(file_handle, str):
        f = open(file_handle, **kwargs)
        opened = True
    elif Path is not None and isinstance(file_handle, Path):
        f = file_handle.open(**kwargs)
        opened = True
    else:
        f = file_handle

    try:
        yield f
    finally:
        if opened:
            f.close()



from __future__ import annotations

__all__ = [
    "convergence_boxplot",
    "location_summary",
    "trumpet_diagram",
]

from os import PathLike
from typing import Callable, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.ticker import FuncFormatter


def convergence_boxplot(
    targets: pd.DataFrame,
    results: pd.DataFrame,
    filter_func: Callable[[pd.Series], pd.Series],
    *,
    adjust_target: bool = True,
    percentage: bool = True,
    band: Tuple[float, float] = None,
    simple_labels: bool = True,
    ax: Axes = None,
    fp: Union[str, PathLike] = None,
    title: str = None,
) -> Axes:
    """Measures convergence of constrained location-choice models (such as work-location choice). Can be used to
    produce multiple box plots for different sub-sets of zones, usually based on size.

    Args:
        targets (pandas.DataFrame):
        results (pandas.DataFrame):
        filter_func (Callable[[pandas.Series], pandas.Series]):
        adjust_target (bool, optional):
        percentage (bool, optional):
        band (Tuple[float, float], optional):
        simple_labels (bool, optional):
        ax (Axes, optional):
        fp (str | PathLike, optional):
        title (str, optional):

    Returns:
        matplotlib.Axes
    """

    assert results.columns.equals(targets.columns)

    columns, filters, n = [], [], 0
    for colname in targets:
        filter_ = filter_func(targets[colname])
        filters.append(filter_)
        n = max(n, filter_.sum())
        columns.append(colname)

    unlabelled_zones = np.full([n, len(columns)], np.nan, dtype=np.float64)
    model_sums, target_sums = [], []

    for i, (colname, filter_) in enumerate(zip(columns, filters)):
        m = filter_.sum()

        model_vector = results.loc[filter_, colname]
        target_vector = targets.loc[filter_, colname]

        if adjust_target and target_vector.sum() > 0:
            factor = model_vector.sum() / target_vector.sum()
            target_vector = target_vector * factor

        model_sums.append(model_vector.sum())
        target_sums.append(target_vector.sum())

        err = model_vector - target_vector
        if percentage:
            err /= target_vector

        unlabelled_zones[:m, i] = err.values

    if not simple_labels:
        columns = [
            "{}\n{} workers\n{} jobs\n{} zones".format(c, model_sums[i], int(target_sums[i]), filters[i].sum())
            for i, c in enumerate(columns)
        ]
    unlabelled_zones = pd.DataFrame(unlabelled_zones, columns=columns)

    with np.errstate(invalid="ignore"):
        ax = unlabelled_zones.plot.box(ax=ax, figsize=[12, 6])
        ax.axhline(0)

        if percentage:
            ax.yaxis.set_major_formatter(FuncFormatter(lambda x, pos: "{}%".format(np.round(x, 2) * 100)))
            ax.set_ylabel("Relative error ((Model - Target) / Target)")
        else:
            ax.set_ylabel("Error (Model - Target)")

        xlabel = "Occupation & Employment Status"
        if adjust_target:
            xlabel += "\n(Targets adjusted to match model totals)"
        ax.set_xlabel(xlabel)

        if band is not None:
            lower, upper = band
            ax.axhline(lower, color="black", linewidth=1, alpha=0.5)
            ax.axhline(upper, color="black", linewidth=1, alpha=0.5)

        if title:
            ax.set_title(title)

        if fp is not None:
            plt.savefig(str(fp))

        return ax


def location_summary(
    model: pd.DataFrame,
    target: pd.DataFrame,
    ensemble_names: pd.Series,
    *,
    title: str = "",
    fp: Union[str, PathLike] = None,
    dpi: int = 150,
    district_name: str = "Ensemble",
) -> Axes:
    """Creates a compound plot showing total attractions to specified locations

    Args:
        model (pandas.DataFrame):
        target (pandas.DataFrame):
        ensemble_names (pandas.Series):
        title (str, optional):
        fp (str | PathLike, optional):
        dpi (int, optional):
        district_name (str, optional):

    Returns:
        matplotlib.Axes
    """

    fig, ax = plt.subplots(1, 3, figsize=[16, 8], gridspec_kw={"width_ratios": [4, 2, 2]})

    model_col = model.reindex(ensemble_names.index, fill_value=0)
    target_col = target.reindex(ensemble_names.index, fill_value=0)

    factor = model_col.sum() / target_col.sum()
    target_col = target_col * factor

    df = pd.DataFrame({"Model": model_col, "Target": target_col})
    df.index = ensemble_names
    df.index.name = district_name
    sub_ax = ax[0]
    df.plot.barh(ax=sub_ax)
    sub_ax.set_title(title)
    sub_ax.invert_yaxis()

    short_labels = np.arange(1, len(ensemble_names) + 1)
    short_labels[-1] = 99

    sub_ax = ax[1]
    diff = model_col - target_col
    diff.index = short_labels

    colours = pd.Series("grey", index=diff.index)
    colours.loc[diff >= 2000] = "skyblue"
    colours.loc[diff >= 5000] = "dodgerblue"
    colours.loc[diff <= -2000] = "lightsalmon"
    colours.loc[diff <= -5000] = "tomato"

    diff.plot.barh(ax=sub_ax, color=colours)
    sub_ax.set_title("Error (Model - Target)")
    sub_ax.invert_yaxis()

    sub_ax = ax[2]
    perc_diff = diff / target_col
    perc_diff.index = short_labels

    colours = pd.Series("grey", index=perc_diff.index)
    colours.loc[perc_diff >= 0.1] = "skyblue"
    colours.loc[perc_diff >= 0.25] = "dodgerblue"
    colours.loc[perc_diff <= -0.1] = "lightsalmon"
    colours.loc[perc_diff <= -0.25] = "tomato"

    perc_diff.plot.barh(ax=sub_ax, color=colours)
    sub_ax.set_title(" % Error (Model - Target) / Target")
    sub_ax.invert_yaxis()

    plt.tight_layout()

    if fp is not None:
        fig.savefig(str(fp), dpi=dpi)

    return ax


def trumpet_diagram(
    counts: pd.Series,
    model_volume: pd.Series,
    *,
    categories: Union[pd.Series, List[pd.Series]] = None,
    category_colours: Dict[Union[str, tuple], str] = None,
    category_markers: Dict[Union[str, tuple], str] = None,
    label_format: str = None,
    title: str = "",
    y_bounds: Tuple[float, float] = (-2, 2),
    ax: Axes = None,
    x_label: str = "Count volume",
    legend: bool = True,
    **kwargs,
) -> Axes:
    """Plots an auto volumes "trumpet" diagram of relative error vs. target count, and will draw min/max error curves
    based on FHWA guidelines. Can be used to plot different categories of count locations.

    Args:
        counts (pandas.Series): Target counts. Each item represents a different count location. Index does not need to
            be unique.
        model_volume (pandas.Series): Modelled volumes for each location. The index must match the counts Series.
        categories (pandas.Series | List[pandas.Series], optional): Defaults to ``None``. Optional classification
            of each count location. Must match the index of the count Series. Can be provided as a List of Series (which
            all must match the count index) to enable tuple-based categorization.
        category_colours (Dict[str | tuple, str], optional): Defaults to ``None``. Mapping of each category to a
            colour, specified as a hex string. Only used when categories are provided. Missing categories revert to
            ``None``, using the default colour for the style.
        category_markers (Dict[str | tuple, str], optional): Defaults to ``None``. Mapping of each category to a
            matplotlib marker string (see https://matplotlib.org/api/markers_api.html for options). Only used when
            categories are provided. Missing categories revert to ``None``, using the default marker for the style.
        label_format (str, optional): Defaults to ``None``. Used to convert category values (especially tuples) into
            readable strings for the plot legend. The relevant line of code is
            ``current_label = label_format % category_key``. Only used when categories are provided.
        title (str, optional): The title to set on the plot.
        y_bounds (tuple, optional): Defaults to ``(-2, 2)``, or (-200%, 200%). Limit of the Y-Axis. This is needed
            because relative errors can be very high close to the y-intercept of the plot.
        ax (matplotlib.Axes, optional): Defaults to ``None``. Sub-Axes to add this plot to, if using ``subplots()``.
        x_label (str, optional): Defaults to ``'Count volume'``. Label to use for the X-axis. The Y-axis is always
            "Relative Error"
        legend (bool, optional): Defaults to ``True``. Flag to add a legend.
        **kwargs: Additional kwargs to pass to ``DataFrame.plot.scatter()``

    Returns:
        matplotlib.Axes:
            The Axes object generated from the plot. For most use cases, this is not really needed.
    """

    assert model_volume.index.equals(counts.index)

    n_categories = 0
    if categories is not None:
        if isinstance(categories, list):
            for s in categories:
                assert s.index.equals(model_volume.index)
            if label_format is None:
                label_format = "-".join(["%s"] * len(categories))
            categories = pd.MultiIndex.from_arrays(categories)
            n_categories = len(categories.unique())
        else:
            assert categories.index.equals(model_volume.index)
            n_categories = categories.nunique()

        if category_colours is None:
            category_colours = {}
        if category_markers is None:
            category_markers = {}
    if label_format is None:
        label_format = "%s"

    df = pd.DataFrame({"Model Volume": model_volume, "Count Volume": counts})
    df["Error"] = df["Model Volume"] - df["Count Volume"]
    df["% Error"] = df["Error"] / df["Count Volume"]

    if n_categories > 1:
        for category_key, subset in df.groupby(categories):
            current_label = label_format % category_key
            current_color = category_colours[category_key] if category_key in category_colours else None
            current_marker = category_markers[category_key] if category_key in category_markers else None

            ax = subset.plot.scatter(
                x="Count Volume",
                y="% Error",
                ax=ax,
                c=current_color,
                marker=current_marker,
                label=current_label,
                **kwargs,
            )
    else:
        ax = df.plot.scatter(x="Count Volume", y="% Error", ax=ax, **kwargs)

    top = counts.max() * 1.05  # Add 5% to the top, to give some visual room on the right side
    xs = np.arange(1, top, 10)
    pos_ys = (-13.7722 + (555.1382 * xs**-0.26025)) / 100.0
    neg_ys = pos_ys * -1

    ax.plot(xs, np.zeros(len(xs)), color="black")
    ax.plot(xs, pos_ys, color="red", linewidth=1, zorder=1)
    ax.plot(xs, neg_ys, color="red", linewidth=1, zorder=1)

    ax.set_xlim(0, top)

    bottom, top = y_bounds
    ax.set_ylim(bottom, top)
    ax.set_yticks(np.arange(bottom, top, 0.25))

    ax.set_title(title)
    ax.set_ylabel("Relative Error")
    ax.set_xlabel(x_label)
    if legend:
        ax.legend()

    return ax



__all__ = [
    "align_categories",
    "is_identifier",
    "reindex_series",
    "sort_nicely",
    "sum_df_sequence",
    "read_feat",
    "read_table_calculator_result_attributes",
    "read_choice_model_component_tables",
    "read_utility_expression_table",
    "read_calibration_target_table",
    "coerce_matrix",
    "expand_array",
    "open_file",
    "read_fortran_rectangle",
    "read_fortran_square",
    "to_fortran",
    "peek_mdf",
    "read_emx",
    "read_mdf",
    "to_emx",
    "to_mdf",
    "read_nwp_base_network",
    "read_nwp_exatts_list",
    "read_nwp_link_attributes",
    "read_nwp_node_attributes",
    "read_nwp_traffic_results",
    "read_nwp_traffic_results_at_countpost",
    "read_nwp_transit_line_attributes",
    "read_nwp_transit_network",
    "read_nwp_transit_result_summary",
    "read_nwp_transit_segment_results",
    "read_nwp_transit_station_results",
    "read_nwp_transit_vehicles",
    "read_omx",
    "to_omx",
    "aggregate_matrix",
    "disaggregate_matrix",
    "fast_stack",
    "fast_unstack",
    "matrix_balancing_1d",
    "matrix_balancing_2d",
    "matrix_bucket_rounding",
    "split_zone_in_matrix",
    "distance_array",
    "distance_matrix",
    "tlfd",
    "generate_time_bin_labels",
]

from .general import align_categories, is_identifier, reindex_series, sort_nicely, sum_df_sequence
from .io import (
    coerce_matrix,
    expand_array,
    open_file,
    peek_mdf,
    read_calibration_target_table,
    read_choice_model_component_tables,
    read_emx,
    read_feat,
    read_fortran_rectangle,
    read_fortran_square,
    read_mdf,
    read_nwp_base_network,
    read_nwp_exatts_list,
    read_nwp_link_attributes,
    read_nwp_node_attributes,
    read_nwp_traffic_results,
    read_nwp_traffic_results_at_countpost,
    read_nwp_transit_line_attributes,
    read_nwp_transit_network,
    read_nwp_transit_result_summary,
    read_nwp_transit_segment_results,
    read_nwp_transit_station_results,
    read_nwp_transit_vehicles,
    read_omx,
    read_table_calculator_result_attributes,
    read_utility_expression_table,
    to_emx,
    to_fortran,
    to_mdf,
    to_omx,
)
from .matrices import (
    aggregate_matrix,
    disaggregate_matrix,
    fast_stack,
    fast_unstack,
    matrix_balancing_1d,
    matrix_balancing_2d,
    matrix_bucket_rounding,
    split_zone_in_matrix,
)
from .modelling import distance_array, distance_matrix, tlfd
from .time import generate_time_bin_labels



from __future__ import annotations

__all__ = [
    "read_feat",
    "read_table_calculator_result_attributes",
    "read_choice_model_component_tables",
    "read_utility_expression_table",
    "read_calibration_target_table",
]

from json import loads
from os import PathLike
from typing import Any, Dict, Tuple, Union

import pandas as pd

try:
    import h5py
except ImportError:
    h5py = None

try:
    from cityphi.feature import Feature
    from inro.emme.agent import Table
except (ImportError, RuntimeError):
    Feature = None
    Table = None


# region Features


if (h5py is None) and (Feature is None) and (Table is None):
    # If neither h5py or OpenPaths Python API is available, gracefully disable the function

    def read_feat(*args, **kwargs):
        raise NotImplementedError()

elif (Feature is None) and (Table is None):
    # If OpenPaths Python API is not available, use h5py to read a basic version of the feature file

    def read_feat(file: Union[str, PathLike]) -> pd.DataFrame:
        """Reads tabular data stored in a feature file produced by OpenPaths applications. Uses h5py.

        Args:
            file (str | PathLike): The file to read

        Returns:
            pd.DataFrame
        """
        retval = []
        with h5py.File(str(file), "r") as f:
            atts = loads(f["attributes"]["data"][()].decode())
            for i, att in enumerate(atts):
                col_name = f"col_{i}"
                if att["dtype"] == "string":
                    v = loads(f["attributes"][col_name][()].decode())
                elif att["is_list"]:
                    continue  # TODO
                else:
                    v = f["attributes"][col_name][()]
                s = pd.Series(v, name=att["name"])
                retval.append(s)

        retval = pd.concat(retval, axis=1)
        if "feature_id" in retval:
            retval.drop("feature_id", axis=1, inplace=True)

        return retval

else:
    # If OpenPaths Python API is available, use it to read a full-featured version of the feature file

    def read_feat(file: Union[str, PathLike]) -> pd.DataFrame:
        """Reads tabular data stored in a feature file produced by OpenPaths applications. Uses the EMME Python API.

        Args:
            file (str | PathLike): The file to read

        Returns:
            pd.DataFrame
        """
        return Table(Feature.from_feature_file(str(file))).to_dataframe()


# endregion

# region Model Package


def _parse_version(ver_str: str) -> Tuple[int, int, int]:
    return tuple([int(v) for v in ver_str.split(".")])


def read_calibration_target_table(model_package_dict: Dict[str, Any]) -> pd.DataFrame:
    """Reads the calibration target table from an AGENT model package specification

    Args:
        model_package_dict (dict): The AGENT model package to parse, as a Python dictionary

    Returns:
        pd.DataFrame
    """
    if _parse_version(model_package_dict["version"]) >= _parse_version("1.3.16"):
        spec = model_package_dict["calibration_targets_table"]
        attribute_info = pd.DataFrame(spec["attribute_info"]).set_index("name")
        component_data = [el.split(";") for el in spec["data"]]
        df = pd.DataFrame.from_records(component_data, columns=attribute_info.index.tolist())
    else:
        df = pd.DataFrame.from_records(model_package_dict["calibration_targets"])
    df["target_min_value"] = pd.to_numeric(df["target_min_value"])
    df["target_max_value"] = pd.to_numeric(df["target_max_value"])
    df["importance"] = pd.to_numeric(df["importance"])
    df["min_calibrated_value"] = pd.to_numeric(df["min_calibrated_value"])
    df["max_calibrated_value"] = pd.to_numeric(df["max_calibrated_value"])
    df["value"] = pd.to_numeric(df["value"])

    return df.set_index("name")


def read_choice_model_component_tables(model_step_dict: Dict[str, Any]) -> pd.DataFrame:
    """Reads the choice components from an AGENT choice model step

    Args:
        model_step_dict (dict): The choice component model step to read in an AGENT model package spec, as a dictionary

    Returns:
        pd.DataFrame
    """
    if model_step_dict["procedure_type"] != "CHOICE_MODEL":
        raise ValueError(f"Model step `{model_step_dict['name']}` is not a choice model")
    choice_components = []
    for i, spec in enumerate(model_step_dict["choice_components"]):
        df = read_utility_expression_table(spec)
        df.insert(0, "choice_component_id", i)
        df.insert(1, "choice_component_name", spec["name"])
        choice_components.append(df)
    choice_components: pd.DataFrame = pd.concat(choice_components, axis=0, ignore_index=True)

    return choice_components


def read_table_calculator_result_attributes(model_step_dict: Dict[str, Any]) -> pd.DataFrame:
    """Reads the result attributes from an AGENT table calculator model step

    Args:
        model_step_dict (dict): The table calculator model step to read in an AGENT model package spec, as a dictionary

    Returns:
        pd.DataFrame
    """
    if model_step_dict["procedure_type"] != "TABLE_CALCULATOR":
        raise ValueError(f"Model step `{model_step_dict['name']}` is not a table calculator")
    df = pd.DataFrame(model_step_dict["result_attributes"])

    return df


def read_utility_expression_table(choice_component_dict: dict) -> pd.DataFrame:
    """Reads the utility expression table from an AGENT model step choice component

    Args:
        choice_component_dict (dict): The choice component of a model step in an AGENT model package spec, as a
            dictionary

    Returns:
        pd.DataFrame
    """
    attribute_info = pd.DataFrame(choice_component_dict["utility_expression_table"]["attribute_info"]).set_index("name")
    component_data = [el.split(";") for el in choice_component_dict["utility_expression_table"]["data"]]
    df = pd.DataFrame.from_records(component_data, columns=attribute_info.index.tolist())
    if not df.empty:
        # Handle wide format
        if choice_component_dict["utility_specification_type"] == "wide":
            df.set_index(["description", "agent_filter", "agent_expression"], inplace=True)
            df.columns.name = "alternative_filter"
            df = df.stack().to_frame("coefficient").reset_index()
            df["alternative_filter"] = df["alternative_filter"].map(attribute_info["description"])
            df["alternative_expression"] = ""

        # Finalize
        if "agent_filter" not in df:
            df["agent_filter"] = ""
        if "agent_expression" not in df:
            df["agent_expression"] = ""
        df = df[
            [
                "description",
                "alternative_filter",
                "alternative_expression",
                "agent_filter",
                "agent_expression",
                "coefficient",
            ]
        ].copy()

    return df


# endregion



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import sys
from pathlib import Path

sys.path.insert(0, Path('../../src').as_posix())

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from datetime import datetime

import wsp_balsa

project = 'wsp-balsa'
author = 'WSP Canada Inc.'
copyright = f'{datetime.today().year}, {author}'

# The short X.Y version
version = ".".join([str(v) for v in wsp_balsa.__version_tuple__[:3]])

# The full version, including alpha/beta/rc tags
release = wsp_balsa.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.githubpages',
    'sphinx_favicon'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_logo = "_static/logo.png"

html_theme_options = {
    'collapse_navigation': True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/wsp-sag/balsa",
            "icon": "fa-brands fa-square-github",
        }
    ]
}

favicons = [
    "favicon-16x16.png",
    "favicon-32x32.png",
    {"rel": "apple-touch-icon", "href": "apple-touch-icon.png"}
]



from __future__ import annotations

__all__ = [
    "aggregate_matrix",
    "disaggregate_matrix",
    "fast_stack",
    "fast_unstack",
    "matrix_balancing_1d",
    "matrix_balancing_2d",
    "matrix_bucket_rounding",
    "split_zone_in_matrix",
]

from multiprocessing import cpu_count
from typing import Callable, Iterable, List, Tuple, Union
from warnings import warn

import numexpr as ne
import numpy as np
import pandas as pd
from numpy.typing import NDArray

try:
    from numba import njit, prange
except ImportError:

    def njit(*args, **kwargs):
        def decorator(func):
            return func

        return decorator

    prange = range

EPS = 1.0e-7


def matrix_balancing_1d(
    m: NDArray,
    a: NDArray,
    axis: int,
) -> NDArray:
    """Balances a matrix using a single constraint.

    Args:
        m (NDArray): The matrix (a 2-dimensional ndarray) to be balanced
        a (NDArray): The totals vector (a 1-dimensional ndarray) constraint
        axis (int): Direction to constrain (0 = along columns, 1 = along rows)

    Return:
        NDArray: A balanced matrix
    """

    assert axis in [0, 1], "axis must be either 0 or 1"
    assert m.ndim == 2, "`m` must be a two-dimensional matrix"
    assert a.ndim == 1, "`a` must be an one-dimensional vector"
    assert np.all(m.shape[axis] == a.shape[0]), "axis %d of matrices 'm' and 'a' must be the same." % axis

    return _balance(m, a, axis)


def matrix_balancing_2d(
    m: Union[NDArray, pd.DataFrame],
    a: NDArray,
    b: NDArray,
    *,
    totals_to_use: str = "raise",
    max_iterations: int = 1000,
    rel_error: float = 0.0001,
    n_threads: int = 1,
) -> Tuple[Union[NDArray, pd.DataFrame], float, int]:
    """Balances a two-dimensional matrix using iterative proportional fitting.

    Args:
        m (NDArray | pandas.DataFrame): The matrix (a 2-dimensional ndarray) to be balanced. If a DataFrame
            is supplied, the output will be returned as a DataFrame.
        a (NDArray): The row totals (a 1-dimensional ndarray) to use for balancing
        b (NDArray): The column totals (a 1-dimensional ndarray) to use for balancing
        totals_to_use (str, optional): Defaults to ``'raise'``. Describes how to scale the row and column totals if
            their sums do not match. Must be one of ['rows', 'columns', 'average', 'raise'].
            - rows: scales the columns totals so that their sums matches the row totals
            - columns: scales the row totals so that their sums matches the column totals
            - average: scales both row and column totals to the average value of their sums
            - raise: raises an Exception if the sums of the row and column totals do not match
        max_iterations (int, optional): Defaults to ``1000``. Maximum number of iterations
        rel_error (float, optional): Defaults to ``1.0E-4``. Relative error stopping criteria
        n_threads (int, optional): Defaults to ``1``. Number of processors for parallel computation. (Not used)

    Return:
        Tuple[NDArray | pandas.DataFrame, float, int]: The balanced matrix, residual, and n_iterations
    """
    max_iterations = int(max_iterations)
    n_threads = int(n_threads)

    # Test if matrix is Pandas DataFrame
    data_type = ""
    m_pd = None
    if isinstance(m, pd.DataFrame):
        data_type = "pd"
        m_pd = m
        m = m_pd.values

    if isinstance(a, pd.Series) or isinstance(a, pd.DataFrame):
        a = a.values
    if isinstance(b, pd.Series) or isinstance(b, pd.DataFrame):
        b = b.values

    # ##################################################################################
    # Validations:
    #   - m is an MxM square matrix, a and b are vectors of size M
    #   - totals_to_use is one of ['rows', 'columns', 'average']
    #   - the max_iterations is a +'ve integer
    #   - rel_error is a +'ve float between 0 and 1
    #   - the n_threads is a +'ve integer between 1 and the number of available processors
    # ##################################################################################
    valid_totals_to_use = ["rows", "columns", "average", "raise"]
    assert m.ndim == 2 and m.shape[0] == m.shape[1], "m must be a two-dimensional square matrix"
    assert a.ndim == 1 and a.shape[0] == m.shape[0], (
        "'a' must be a one-dimensional array, whose size matches that of 'm'"
    )
    assert b.ndim == 1 and b.shape[0] == m.shape[0], (
        "'a' must be a one-dimensional array, whose size matches that of 'm'"
    )
    assert totals_to_use in valid_totals_to_use, "totals_to_use must be one of %s" % valid_totals_to_use
    assert max_iterations >= 1, "max_iterations must be integer >= 1"
    assert 0 < rel_error < 1.0, "rel_error must be float between 0.0 and 1.0"
    assert 1 <= n_threads <= cpu_count(), (
        "n_threads must be integer between 1 and the number of processors (%d) " % cpu_count()
    )
    if n_threads > 1:
        raise NotImplementedError("Multiprocessing capability is not implemented yet.")

    # Scale row and column totals, if required
    a_sum = a.sum()
    b_sum = b.sum()
    if not np.isclose(a_sum, b_sum):
        if totals_to_use == "rows":
            b = np.multiply(b, a_sum / b_sum)
        elif totals_to_use == "columns":
            a = np.multiply(a, b_sum / a_sum)
        elif totals_to_use == "average":
            avg_sum = 0.5 * (a_sum + b_sum)
            a = np.multiply(a, avg_sum / a_sum)
            b = np.multiply(b, avg_sum / b_sum)
        else:
            raise RuntimeError("a and b vector totals do not match.")

    initial_error = _calc_error(m, a, b)
    err = 1.0
    i = 0
    while err > rel_error:
        if i > max_iterations:
            warn("Matrix balancing did not converge", RuntimeWarning)
            break
        m = _balance(m, a, 1)
        m = _balance(m, b, 0)
        err = _calc_error(m, a, b) / initial_error
        i += 1

    if data_type == "pd":
        new_df = pd.DataFrame(m, index=m_pd.index, columns=m_pd.columns)
        return new_df, err, i
    else:
        return m, err, i


def _balance(matrix: NDArray, tot: NDArray, axis: int) -> NDArray:
    """Balances a matrix using a single constraint.

    Args:
        matrix (NDArray): The matrix to be balanced
        tot (NDArray): The totals constraint
        axis (int): Direction to constrain (0 = along columns, 1 = along rows)

    Return:
        NDArray: The balanced matrix
    """
    sc = tot / (matrix.sum(axis) + EPS)
    sc = np.nan_to_num(sc)  # replace divide by 0 errors from the prev. line
    if axis:  # along rows
        matrix = np.multiply(matrix.T, sc).T
    else:  # along columns
        matrix = np.multiply(matrix, sc)
    return matrix


def _calc_error(m, a, b):
    row_sum = np.absolute(a - m.sum(1)).sum()
    col_sum = np.absolute(b - m.sum(0)).sum()
    return row_sum + col_sum


@njit(cache=False, parallel=True)
def _nbf_bucket_round(a_, decimals=0):
    a = a_.ravel()
    b = np.copy(a)

    residual = 0
    for i in prange(0, len(b)):
        b[i] = np.round(a[i] + residual, decimals)
        residual += a[i] - b[i]

    return b.reshape(a_.shape)


def matrix_bucket_rounding(
    m: Union[NDArray, pd.DataFrame],
    *,
    decimals: int = 0,
) -> Union[NDArray, pd.DataFrame]:
    """Bucket rounds to the given number of decimals.

    Args:
        m (NDArray | pandas.DataFrame): The matrix to be rounded
        decimals (int, optional): Defaults to ``0``. Number of decimal places to round to. If decimals is negative, it
            specifies the number of positions to the left of the decimal point.

    Return:
        NDArray | pandas.DataFrame: The rounded matrix
    """

    # Test if matrix is Pandas DataFrame
    data_type = ""
    m_pd = None
    if isinstance(m, pd.DataFrame):
        data_type = "pd"
        m_pd = m
        m = m_pd.values

    decimals = int(decimals)

    # I really can't think of a way to vectorize bucket rounding, so here goes the slow for loop
    b = _nbf_bucket_round(m, decimals)

    if decimals <= 0:
        b = b.astype(np.int32)

    if data_type == "pd":
        new_df = pd.DataFrame(b.reshape(m.shape), index=m_pd.index, columns=m_pd.columns)
        return new_df
    else:
        return b.reshape(m.shape)


def split_zone_in_matrix(
    base_matrix: pd.DataFrame,
    old_zone: int,
    new_zones: List[int],
    proportions: List[float],
) -> pd.DataFrame:
    """Takes a zone in a matrix (as a DataFrame) and splits it into several new zones, prorating affected cells by a
    vector of proportions (one value for each new zone). The old zone is removed.

    Args:
        base_matrix (pandas.DataFrame): The matrix to re-shape
        old_zone (int): The original zone to split
        new_zones (List[int]): The list of new zones to add
        proportions (List[float]): The proportions to split the original zone to. The list must be the same length as
            ``new_zones`` and sum to 1.0

    Returns:
        pandas.DataFrame: The re-shaped matrix
    """

    assert isinstance(base_matrix, pd.DataFrame), "Base matrix must be a DataFrame"

    old_zone = int(old_zone)
    new_zones = np.array(new_zones, dtype=np.int32)
    proportions = np.array(proportions, dtype=np.float64)

    assert len(new_zones) == len(proportions), "Proportion array must be the same length as the new zone array"
    assert len(new_zones.shape) == 1, "New zones must be a vector"
    assert base_matrix.index.equals(base_matrix.columns), "DataFrame is not a matrix"
    assert np.isclose(proportions.sum(), 1.0), "Proportions must sum to 1.0 "

    n_new_zones = len(new_zones)

    intersection_index = base_matrix.index.drop(old_zone)
    new_index = intersection_index
    for z in new_zones:
        new_index = new_index.insert(-1, z)
    new_index = pd.Index(sorted(new_index))

    new_matrix = pd.DataFrame(0, index=new_index, columns=new_index, dtype=base_matrix.dtypes.iat[0])

    # 1. Copy over the values from the regions of the matrix not being updated
    new_matrix.loc[intersection_index, intersection_index] = base_matrix

    # 2. Prorate the row corresponding to the dropped zone
    # This section (and the next) works with the underlying Numpy arrays, since they handle
    # broadcasting better than Pandas does
    original_row = base_matrix.loc[old_zone, intersection_index]
    original_row = original_row.values[:]  # Make a shallow copy to preserve shape of the original data
    original_row.shape = 1, len(intersection_index)
    proportions.shape = n_new_zones, 1
    result = pd.DataFrame(original_row * proportions, index=new_zones, columns=intersection_index)
    new_matrix.loc[result.index, result.columns] = result

    # 3. Proprate the column corresponding to the dropped zone
    original_column = base_matrix.loc[intersection_index, old_zone]
    original_column = original_column.values[:]
    original_column.shape = len(intersection_index), 1
    proportions.shape = 1, n_new_zones
    result = pd.DataFrame(original_column * proportions, index=intersection_index, columns=new_zones)
    new_matrix.loc[result.index, result.columns] = result

    # 4. Expand the old intrazonal
    proportions_copy = proportions[:, :]
    proportions_copy.shape = 1, n_new_zones
    proportions.shape = n_new_zones, 1

    intrzonal_matrix = proportions * proportions_copy
    intrazonal_scalar = base_matrix.at[old_zone, old_zone]

    result = pd.DataFrame(intrazonal_scalar * intrzonal_matrix, index=new_zones, columns=new_zones)
    new_matrix.loc[result.index, result.columns] = result

    return new_matrix


def aggregate_matrix(
    matrix: Union[pd.DataFrame, pd.Series],
    *,
    groups: Union[pd.Series, NDArray] = None,
    row_groups: Union[pd.Series, NDArray] = None,
    col_groups: Union[pd.Series, NDArray] = None,
    aggfunc: Callable[[Iterable[Union[int, float]]], Union[int, float]] = np.sum,
) -> Union[pd.DataFrame, pd.Series]:
    """Aggregates a matrix based on mappings provided for each axis, using a specified aggregation function.

    Args:
        matrix (pandas.DataFrame | pandas.Series): Matrix data to aggregate. DataFrames and Series with 2-level
            indices are supported
        groups (pandas.Series | NDArray, optional): Syntactic sugar to specify both row_groups and
            col_groups to use the same grouping series.
        row_groups (pandas.Series | NDArray, optional): Groups for the rows. If aggregating a DataFrame,
            this must match the index of the matrix. For a "tall" matrix, this series can match either the "full" index
            of the series, or it can match the first level of the matrix (it would be the same as if aggregating a
            DataFrame). Alternatively, an array can be provided, but it must be the same length as the DataFrame's
            index, or the full length of the Series.
        col_groups (pandas.Series | NDArray, optional): Groups for the columns. If aggregating a DataFrame,
            this must match the columns of the matrix. For a "tall" matrix, this series can match either the "full"
            index of the series, or it can match the second level of the matrix (it would be the same as if aggregating
            a DataFrame). Alternatively, an array can be provided, but it must be the same length as the DataFrame's
            columns, or the full length of the Series.
        aggfunc: The aggregation function to use. Default is np.sum.

    Returns:
        pandas.Series or pandas.DataFrame:
            The aggregated matrix, in the same type as was provided, e.g. Series -> Series, DataFrame -> DataFrame.

    Example:

        matrix:

        +-------+---+---+---+---+---+---+---+
        |       | 1 | 2 | 3 | 4 | 5 | 6 | 7 |
        +=======+===+===+===+===+===+===+===+
        | **1** | 2 | 1 | 9 | 6 | 7 | 8 | 5 |
        +-------+---+---+---+---+---+---+---+
        | **2** | 4 | 1 | 1 | 4 | 8 | 7 | 6 |
        +-------+---+---+---+---+---+---+---+
        | **3** | 5 | 8 | 5 | 3 | 5 | 9 | 4 |
        +-------+---+---+---+---+---+---+---+
        | **4** | 1 | 1 | 2 | 9 | 4 | 9 | 9 |
        +-------+---+---+---+---+---+---+---+
        | **5** | 6 | 3 | 4 | 6 | 9 | 9 | 3 |
        +-------+---+---+---+---+---+---+---+
        | **6** | 7 | 2 | 5 | 8 | 2 | 5 | 9 |
        +-------+---+---+---+---+---+---+---+
        | **7** | 3 | 1 | 8 | 6 | 3 | 5 | 6 |
        +-------+---+---+---+---+---+---+---+

        groups:

        +-------+---+
        | **1** | A |
        +-------+---+
        | **2** | B |
        +-------+---+
        | **3** | A |
        +-------+---+
        | **4** | A |
        +-------+---+
        | **5** | C |
        +-------+---+
        | **6** | C |
        +-------+---+
        | **7** | B |
        +-------+---+

        ``new_matrix = aggregate_matrix(matrix, groups=groups)``

        new_matrix:

        +-------+----+----+----+
        |       | A  | B  | C  |
        +=======+====+====+====+
        | **A** | 42 | 28 | 42 |
        +-------+----+----+----+
        | **B** | 26 | 14 | 23 |
        +-------+----+----+----+
        | **C** | 36 | 17 | 25 |
        +-------+----+----+----+

    """
    if groups is not None:
        row_groups = groups
        col_groups = groups

    assert row_groups is not None, "Row groups must be specified"
    assert col_groups is not None, "Column groups must be specified"

    if isinstance(matrix, pd.DataFrame):
        row_groups = _prep_square_index(matrix.index, row_groups)
        col_groups = _prep_square_index(matrix.columns, col_groups)

        return _aggregate_frame(matrix, row_groups, col_groups, aggfunc)
    elif isinstance(matrix, pd.Series):
        assert matrix.index.nlevels == 2

        row_groups, col_groups = _prep_tall_index(matrix.index, row_groups, col_groups)
        return _aggregate_series(matrix, row_groups, col_groups, aggfunc)
    else:
        raise NotImplementedError()


def _prep_tall_index(target_index, row_aggregator, col_aggregator):
    if isinstance(row_aggregator, pd.Series):
        if row_aggregator.index.equals(target_index):
            row_aggregator = row_aggregator.values
        else:
            assert target_index.levels[0].equals(row_aggregator.index)
            reindexed = row_aggregator.reindex(target_index, level=0)
            row_aggregator = reindexed.values
    else:
        assert len(row_aggregator) == len(target_index)
        row_aggregator = np.array(row_aggregator)

    if isinstance(col_aggregator, pd.Series):
        if col_aggregator.index.equals(target_index):
            col_aggregator = col_aggregator.values
        else:
            assert target_index.levels[1].equals(col_aggregator.index)
            reindexed = col_aggregator.reindex(target_index, level=1)
            col_aggregator = reindexed.values
    else:
        assert len(col_aggregator) == len(target_index)
        col_aggregator = np.array(col_aggregator)

    return row_aggregator, col_aggregator


def _prep_square_index(index, aggregator):
    if isinstance(aggregator, pd.Series):
        assert aggregator.index.equals(index)
        return aggregator.values
    else:
        assert len(aggregator) == len(index)
        return np.array(aggregator)


def _aggregate_frame(matrix: pd.DataFrame, row_aggregator, col_aggregator, aggfunc):
    return matrix.groupby(row_aggregator, axis=0).aggregate(aggfunc).groupby(col_aggregator, axis=1).aggregate(aggfunc)


def _aggregate_series(matrix: pd.Series, row_aggregator, col_aggregator, aggfunc):
    return matrix.groupby([row_aggregator, col_aggregator]).aggregate(aggfunc)


def fast_stack(
    frame: pd.DataFrame,
    multi_index: pd.MultiIndex,
    *,
    deep_copy: bool = True,
) -> pd.Series:
    """Performs the same action as ``DataFrame.stack()``, but provides better performance when the target stacked index
    is known beforehand. Useful in converting a lot of matrices from "wide" to "tall" format. The inverse of
    ``fast_unstack()``.

    Notes:
        This function does not check that the entries in the multi_index are compatible with the index and columns of
        the source DataFrame, only that the lengths are compatible. It can therefore be used to assign a whole new set
        of labels to the result.

    Args:
        frame (pandas.DataFrame): The DataFrame to stack.
        multi_index (pandas.MultiIndex): The 2-level MultiIndex known ahead-of-time.
        deep_copy (bool, optional): Defaults to ``True``. A flag indicating if the returned Series should be a view of
            the underlying data (deep_copy=False) or a copy of it (deep_copy=True). A deep copy takes a little longer to
            convert and takes up more memory but preserves the original data of the DataFrame. The default value of True
            is recommended for most uses.

    Returns:
        pandas.Series: The stacked data.
    """

    assert multi_index.nlevels == 2, "Target index must be a MultiIndex with exactly 2 levels"
    assert len(multi_index) == len(frame.index) * len(frame.columns), (
        "Target index and source index and columns do not have compatible lengths"
    )

    array = np.ascontiguousarray(frame.values)
    array = array.copy() if deep_copy else array[:, :]
    array.shape = len(frame.index) * len(frame.columns)

    return pd.Series(array, index=multi_index)


def fast_unstack(
    series: pd.Series,
    index: pd.Index,
    columns: pd.Index,
    *,
    deep_copy: bool = True,
) -> pd.DataFrame:
    """Performs the same action as ``DataFrame.unstack()``, but provides better performance when the target unstacked
    index and columns are known beforehand. Useful in converting a lot of matrices from "tall" to "wide" format. The
    inverse of ``fast_stack()``.

    Notes:
        This function does not check that the entries in index and columns are compatible with the MultiIndex of the
        source Series, only that the lengths are compatible. It can therefore be used to assign a whole new set of
        labels to the result.

    Args:
        series (pandas.Series): The Series with 2-level MultiIndex to unstack
        index (pandas.Index): The row index known ahead-of-time
        columns (pandas.Index): The columns index known ahead-of-time.
        deep_copy (bool): Defaults to ``True``. A flag indicating if the returned DataFrame should be a view of the
            underlying data (deep_copy=False) or a copy of it (deep_copy=True). A deep copy takes a little longer to
            convert and takes up more memory but preserves the original data of the Series. The default value of True is
            recommended for most uses.

    Returns:
        pandas.DataFrame: The unstacked dat
    """

    assert series.index.nlevels == 2, "Source Series must have an index with exactly 2 levels"
    assert len(series) == len(index) * len(columns), (
        "Source index and target index and columns do not have compatible lengths"
    )

    array = series.values.copy() if deep_copy else series.values[:]
    array.shape = len(index), len(columns)

    return pd.DataFrame(array, index=index, columns=columns)


def _check_disaggregation_input(mapping: pd.Series, proportions: pd.Series) -> NDArray:
    assert mapping is not None
    assert proportions is not None
    assert mapping.index.equals(proportions.index)

    # Force proportions to sum to 1 by dividing by the total in each parent
    parent_totals = (
        proportions.groupby(mapping)  # Group the proportions by parent zones
        .sum()  # Sum the total for each parent
        .reindex(mapping)  # Reindex for all child zones
        .values  # Get the ndarray to avoid index alignment problems
    )

    return proportions.values / parent_totals


def disaggregate_matrix(
    matrix: pd.DataFrame,
    *,
    mapping: pd.Series = None,
    proportions: pd.Series = None,
    row_mapping: pd.Series = None,
    row_proportions: pd.Series = None,
    col_mapping: pd.Series = None,
    col_proportions: pd.Series = None,
) -> pd.DataFrame:
    """Split multiple rows and columns in a matrix all at once. The cells in the matrix MUST be numeric, but the row
    and column labels do not.

    Args:
        matrix (pandas.DataFrame): The input matrix to disaggregate
        mapping (pandas.Series, optional): Dict-like Series of "New label" : "Old label". Sets both the row_mapping and
            col_mapping variables if provided (resulting in a square matrix).
        proportions (pandas.Series, optional): Dict-like Series of "New label": "Proportion of old label". Its index
            must match the index of the mapping argument. Sets both the row_proportions and col_proportions arguments
            if provided.
        row_mapping (pandas.Series, optional): Same as mapping, except applied only to the rows.
        row_proportions (pandas.Series, optional): Same as proportions, except applied only to the rows
        col_mapping (pandas.Series, optional): Same as mapping, except applied only to the columns.
        col_proportions (pandas.Series, optional): Same as proportions, except applied only to the columns

    Returns:
        pandas.DataFrame: An expanded DataFrame with the new indices. The new matrix will sum to the same total as the original.

    Examples:

        df:

        +---+----+----+----+
        |   | A  | B  | C  |
        +===+====+====+====+
        | A | 10 | 30 | 20 |
        +---+----+----+----+
        | B | 20 | 10 | 10 |
        +---+----+----+----+
        | C | 30 | 20 | 20 |
        +---+----+----+----+

        correspondence:

        +-----+-----+------+
        | new | old | prop |
        +=====+=====+======+
        | A1  |  A  | 0.25 |
        +-----+-----+------+
        | A2  |  A  | 0.75 |
        +-----+-----+------+
        | B1  |  B  | 0.55 |
        +-----+-----+------+
        | B2  |  B  | 0.45 |
        +-----+-----+------+
        | C1  |  C  | 0.62 |
        +-----+-----+------+
        | C2  |  C  | 0.38 |
        +-----+-----+------+

        ``new_matrix = disaggregate_matrix(df, mapping=correspondence['old'], proportions=correspondence['prop'])``

        new_matrix:

        +-----+-------+-------+--------+--------+-------+-------+
        | new |  A1   | A2    | B1     | B2     | C1    | C2    |
        +=====+=======+=======+========+========+=======+=======+
        |  A1 | 0.625 | 1.875 |  4.125 |  3.375 | 3.100 | 1.900 |
        +-----+-------+-------+--------+--------+-------+-------+
        |  A2 | 1.875 | 5.625 | 12.375 | 10.125 | 9.300 | 5.700 |
        +-----+-------+-------+--------+--------+-------+-------+
        |  B1 | 2.750 | 8.250 |  3.025 |  2.475 | 3.410 | 2.090 |
        +-----+-------+-------+--------+--------+-------+-------+
        |  B2 | 2.250 | 6.750 |  2.475 |  2.025 | 2.790 | 1.710 |
        +-----+-------+-------+--------+--------+-------+-------+
        |  C1 | 4.650 | 13.95 |  6.820 |  5.580 | 7.688 | 4.712 |
        +-----+-------+-------+--------+--------+-------+-------+
        |  C2 | 2.850 |  8.55 |  4.180 |  3.420 | 4.712 | 2.888 |
        +-----+-------+-------+--------+--------+-------+-------+

    """

    # Check that all inputs are specified
    if mapping is not None:
        row_mapping, col_mapping = mapping, mapping
    if proportions is not None:
        row_proportions, col_proportions = proportions, proportions

    row_proportions = _check_disaggregation_input(row_mapping, row_proportions)
    col_proportions = _check_disaggregation_input(col_mapping, col_proportions)

    # Validate inputs
    new_rows = row_mapping.index
    new_cols = col_mapping.index

    # Get raw indexers for NumPy & lookup the value in each parent cell
    row_indexer = matrix.index.get_indexer(row_mapping)[:, np.newaxis]
    col_indexer = matrix.columns.get_indexer(col_mapping)[np.newaxis, :]
    parent_cells = matrix.values[row_indexer, col_indexer]

    # Convert proportions to 2D vectors
    row_proportions = row_proportions[:, np.newaxis]
    col_proportions = col_proportions[np.newaxis, :]

    # Multiply each parent cell by its disaggregation proportion & return
    result_matrix = ne.evaluate("parent_cells * row_proportions * col_proportions")

    result_matrix = pd.DataFrame(result_matrix, index=new_rows, columns=new_cols)
    return result_matrix



from __future__ import annotations

__all__ = [
    "read_fortran_rectangle",
    "read_fortran_square",
    "to_fortran",
]

from io import FileIO
from pathlib import Path
from typing import Iterable, Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from .common import coerce_matrix, expand_array, open_file


def _infer_fortran_zones(n_words):
    """Returns the inverse of n_words = matrix_size * (matrix_size + 1)"""
    n = int(0.5 + ((1 + 4 * n_words) ** 0.5) / 2) - 1
    if n_words != (n * (n + 1)):
        raise RuntimeError("Could not infer a square matrix from file")
    return n


def read_fortran_rectangle(
    file: Union[str, FileIO, Path],
    n_columns: int,
    *,
    zones: Union[int, Iterable[int], pd.Index] = None,
    tall: bool = False,
    reindex_rows: bool = False,
    fill_value: Union[int, float] = None,
) -> Union[NDArray, pd.DataFrame, pd.Series]:
    """Reads a FORTRAN-friendly .bin file (a.k.a. 'simple binary format') which is known to NOT be square. Also works
    with square matrices.

    This file format is an array of 4-bytes, where each row is prefaced by an integer referring to the 1-based
    positional index that FORTRAN uses. The rest of the data are in 4-byte floats. To read this, the number of columns
    present must be known, since the format does not self-specify.

    Args:
        file(str | FileIO | Path): The file to read.
        n_columns (int): The number of columns in the matrix.
        zones (int | Iterable[int] | pandas.Index, optional): Defaults to ``None``. An Index or Iterable will be
            interpreted as the zone labels for the matrix rows and columns; returning a DataFrame or Series (depending
            on `tall`). If an integer is provided, the returned ndarray will be truncated to this 'number of zones'.
        tall (bool, optional): Defaults to ``False``. If true, a 'tall' version of the matrix will be returned.
        reindex_rows (bool, optional): Defaults to ``False``. If true, and zones is an Index, the returned DataFrame
            will be reindexed to fill-in any missing rows.
        fill_value (optional): Defaults to ``None``. The value to pass to ``pandas.reindex()``

    Returns:
        NDArray, pandas.DataFrame or pandas.Series

    Raises:
        ValueError: if the shape is not valid.
    """
    with open_file(file, mode="rb") as reader:
        n_columns = int(n_columns)

        matrix = np.fromfile(reader, dtype=np.float32)
        rows = len(matrix) // (n_columns + 1)
        if len(matrix) != (rows * (n_columns + 1)):
            raise ValueError("Matrix shape is not valid")

        matrix.shape = rows, n_columns + 1

        # Convert binary representation from float to int, then subtract 1 since FORTRAN uses 1-based positional
        # indexing
        row_index = np.frombuffer(matrix[:, 0].tobytes(), dtype=np.int32) - 1
        matrix = matrix[:, 1:]

        if zones is None:
            if tall:
                matrix.shape = matrix.shape[0] * matrix.shape[1]
            return matrix

        if isinstance(zones, (int, np.int_)):
            matrix = matrix[:zones, :zones]

            if tall:
                matrix.shape = zones * zones
            return matrix

        nzones = len(zones)
        matrix = matrix[:nzones, :nzones]
        row_labels = zones.take(row_index[:nzones])
        matrix = pd.DataFrame(matrix, index=row_labels, columns=zones)

        if reindex_rows:
            matrix = matrix.reindex(zones, axis=0, fill_value=fill_value)

        if tall:
            return matrix.stack()
        return matrix


def read_fortran_square(
    file: Union[str, FileIO, Path],
    *,
    zones: Union[int, Iterable[int], pd.Index] = None,
    tall: bool = False,
) -> Union[NDArray, pd.DataFrame, pd.Series]:
    """Reads a FORTRAN-friendly .bin file (a.k.a. 'simple binary format') which is known to be square.

    This file format is an array of 4-bytes, where each row is prefaced by an integer referring to the 1-based
    positional index that FORTRAN uses. The rest of the data are in 4-byte floats. To read this, the number of columns
    present must be known, since the format does not self-specify. This method can infer the shape if it is square.

    Args:
        file (str | FileIO | Path): The file to read.
        zones (int | pandas.Index | Iterable[int], optional): Defaults to ``None``. An Index or Iterable will be
            interpreted as the zone labels for the matrix rows and columns; returning a DataFrame or Series (depending
            on ``tall``). If an integer is provided, the returned ndarray will be truncated to this 'number of zones'.
            Otherwise, the returned ndarray will be size to the maximum number of zone dimensioned by the Emmebank.
        tall (bool, optional): Defaults to ``False``. If True, a 1D data structure will be returned. If ``zone_index``
            is provided, a Series will be returned, otherwise a 1D ndarray.

    Returns:
        NDArray, pandas.DataFrame, or pandas.Series
    """
    with open_file(file, mode="rb") as reader:
        floats = np.fromfile(reader, dtype=np.float32)
        n_words = len(floats)
        matrix_size = _infer_fortran_zones(n_words)
        floats.shape = matrix_size, matrix_size + 1

        data = floats[:, 1:]

        if zones is None:
            if tall:
                n = np.prod(data.shape)
                data.shape = n
                return data
            return data

        if isinstance(zones, (int, np.int_)):
            data = data[:zones, :zones]

            if tall:
                data.shape = zones * zones
                return data
            return data
        elif zones is None:
            return data

        zones = pd.Index(zones)
        n = len(zones)
        data = data[:n, :n]

        matrix = pd.DataFrame(data, index=zones, columns=zones)

        return matrix.stack() if tall else matrix


def to_fortran(
    matrix: Union[NDArray, pd.DataFrame, pd.Series],
    file: Union[str, FileIO, Path],
    *,
    n_columns: int = None,
    min_index: int = 1,
    force_square: bool = True,
) -> None:
    """Writes a FORTRAN-friendly .bin file (a.k.a. 'simple binary format'), in a square format.

    Args:
        matrix (pandas.DataFrame | pandas.Series | NDArray): The matrix to write to disk. If a Series is
            given, it MUST have a MultiIndex with exactly 2 levels to unstack.
        file (str | FileIO | Path): The path or file handler to write to.
        n_columns (int, optional): Defaults to ``None``. Specifies a desired "width" of the matrix file. For example,
            ``n_columns=4000`` on a 3500x3500 matrix will pad the width with 500 extra columns containing 0. If ``None``
            is provided or the value is <= the width of the given matrix, no padding will be performed.
        min_index (int, optional): Defaults to ``1``. The lowest numbered row. Used when slicing matrices
        force_square (bool, optional): Defaults to ``True``.
    """
    if min_index < 1:
        raise ValueError("`min_index` must be >= 1")
    array = coerce_matrix(matrix, force_square=force_square)

    if n_columns is not None and n_columns > array.shape[1]:
        extra_columns = n_columns - array.shape[1]
        array = expand_array(array, extra_columns, axis=1)

    with open_file(file, mode="wb") as writer:
        rows, columns = array.shape
        temp = np.zeros([rows, columns + 1], dtype=np.float32)
        temp[:, 1:] = array

        index = np.arange(min_index, rows + 1, dtype=np.int32)
        # Mask the integer binary representation as floating point
        index_as_float = np.frombuffer(index.tobytes(), dtype=np.float32)
        temp[:, 0] = index_as_float

        temp.tofile(writer)



from __future__ import annotations

__all__ = [
    "LogFormats",
    "ModelLogger",
    "init_root",
    "get_model_logger",
    "log_to_file",
]

import logging
import sys
import traceback as tb
from contextlib import contextmanager
from enum import Enum
from json import dumps as json_to_str
from json import loads as parse_json
from logging import FileHandler, Formatter, Logger, LogRecord
from os import PathLike
from typing import Any, Dict, Generator, Union

try:
    from inro.modeller import logbook_write

    EMME_ENV = True
except (ImportError, RuntimeError):
    logbook_write = None
    EMME_ENV = False


# region Constants

_FMT_STRING = "%(asctime)s %(levelname)s %(name)s {arrow} %(message)s"
_UNC_ARROW = "➔"
_ASCII_ARROW = "->"
_TIP_LEVEL = 21  # Important log info
_REPORT_LEVEL = 19  # Model results
_SUBPROC_LEVEL = 14
_SUBPROC_ERR_LEVEL = 41
_FANCY_FORMAT = "FANCY"
_BASIC_FORMAT = "BASIC"
_JSON_FORMAT = "JSON"


class LogFormats(Enum):
    FANCY = _FANCY_FORMAT
    BASIC = _BASIC_FORMAT
    JSON = _JSON_FORMAT


# endregion


# region Filter Classes


class _RangeFilter(object):
    def __init__(self, low: int, high: int):
        self._low = int(low)
        self._high = int(high)

    def filter(self, record: LogRecord) -> int:
        return int(self._low <= record.levelno <= self._high)


# endregion


# region Formatter Classes


class _SwitchFormatter(Formatter):
    def __init__(self, default_format: Union[Formatter, str], level_formats: Dict[int, Union[Formatter, str]]):
        super(_SwitchFormatter, self).__init__()

        def make_formatter(item: Union[Formatter, str]) -> Formatter:
            return Formatter(item) if isinstance(item, str) else item

        self._default = make_formatter(default_format)
        self._formats = {lvl: make_formatter(f) for lvl, f in level_formats.items()}

    def format(self, record: LogRecord) -> str:
        level = record.levelno
        if level in self._formats:
            return self._formats[level].format(record)
        return self._default.format(record)


class _JsonFormatter(Formatter):
    def format(self, record: LogRecord) -> str:
        keys = ["levelname", "name", "msg", "created", "levelno"]
        to_json = {key: getattr(record, key) for key in keys}
        to_json["asctime"] = self.formatTime(record)
        return json_to_str(to_json)


# endregion


class ModelLogger(Logger):
    def __init__(self, name: str, level: int = logging.NOTSET):
        """ModelLogger extends the standard Python Logger, adding additional statements such as ``.report()``."""
        super(ModelLogger, self).__init__(name, level)
        self._all_load_failures = []

    def report(self, msg, *args, **kwargs):
        """Report useful model statistics or results to the user. Distinct from ``.info()`` which provides status
        information. Printed in green when colours are available."""
        self.log(_REPORT_LEVEL, msg, *args, **kwargs)

    def tip(self, msg, *args, **kwargs):
        """Provide a more significant status statement (e.g. new section of the model). Similar to ``.info()``, but
        more emphasized. Printed in blue when colours are available."""
        self.log(_TIP_LEVEL, msg, *args, **kwargs)

    def subproc_message(self, msg, *args, **kwargs):
        """Report subprocess messages to the user. Distinct from ``.info()`` which provides status information."""
        self.log(_SUBPROC_LEVEL, msg, *args, **kwargs)

    def subproc_err(self, msg, *args, **kwargs):
        """Report subprocess errors to the user. Distinct from ``.error()`` which provides error information."""
        self.log(_SUBPROC_ERR_LEVEL, msg, *args, **kwargs)

    def log_json(self, json_string: str):
        json_dict: Dict[str, Any] = parse_json(json_string)
        record = logging.makeLogRecord(json_dict)
        self.handle(record)

    def load_failure(self, msg: str):
        self._all_load_failures.append(msg)
        self.warning(msg)

    if EMME_ENV:

        def flush_load_failures(self):
            body = "\n".join(self._all_load_failures)
            logbook_write(name="List of all errors in model configuration", value=body)

        def __del__(self):
            self.flush_load_failures()


def _prep_fancy_formatter():
    raw_fmt = _FMT_STRING.format(arrow=_UNC_ARROW)
    fmt_string = str("".join(["\x1b[{colour}m", raw_fmt, "\x1b[0m"]))

    debug_formatter = logging.Formatter(fmt_string.format(colour=37))  # Grey colour
    subproc_formatter = logging.Formatter(fmt_string.format(colour=37))  # Grey colour
    report_formatter = logging.Formatter(fmt_string.format(colour=32))  # Green colour
    info_formatter = logging.Formatter(fmt_string.format(colour=0))  # Default colour
    tip_formatter = logging.Formatter(fmt_string.format(colour=34))  # Blue colour
    warn_formatter = logging.Formatter(fmt_string.format(colour=31))  # Red colour
    error_formatter = logging.Formatter(fmt_string.format(colour=41))  # Red BG colour
    subproc_err_formatter = logging.Formatter(fmt_string.format(colour=41))  # Red BG colour
    critical_formatter = logging.Formatter(fmt_string.format(colour="1m\x1b[41"))  # Bold on red BG

    switch_formatter = _SwitchFormatter(
        raw_fmt,
        {
            logging.INFO: info_formatter,
            logging.WARNING: warn_formatter,
            _TIP_LEVEL: tip_formatter,
            _REPORT_LEVEL: report_formatter,
            logging.DEBUG: debug_formatter,
            logging.ERROR: error_formatter,
            logging.CRITICAL: critical_formatter,
            _SUBPROC_LEVEL: subproc_formatter,
            _SUBPROC_ERR_LEVEL: subproc_err_formatter,
        },
    )

    return switch_formatter


# region Helper Functions


def init_root(
    name: str,
    *,
    stream_format: Union[str, LogFormats] = LogFormats.FANCY,
    log_debug: bool = True,
) -> ModelLogger:
    """Initialize a ModelLogger

    Args:
        root_name (str): The logger's name to use
        stream_format (str | LogFormats, optional): Defaults to ``LogFormats.FANCY``. Logging format options to use.
            Options include ``BASIC``, ``FANCY``, or ``JSON``
        log_debug (bool, optional): Defaults to ``True``. A flag to log debug-level messages
    """

    logging.addLevelName(_SUBPROC_ERR_LEVEL, "SUBPROC_ERR")
    logging.addLevelName(_TIP_LEVEL, "TIP")
    logging.addLevelName(_REPORT_LEVEL, "REPORT")
    logging.addLevelName(_SUBPROC_LEVEL, "SUBPROC")
    logging.setLoggerClass(ModelLogger)

    root_logger: ModelLogger = logging.getLogger(name)  # Will return a ModelLogger based on the previous line
    root_logger.propagate = True
    root_logger.setLevel(1 if log_debug else _REPORT_LEVEL)

    stream_format = LogFormats(stream_format)
    if stream_format == LogFormats.FANCY:
        stdout_formatter = _prep_fancy_formatter()
    elif stream_format == LogFormats.BASIC:
        stdout_formatter = logging.Formatter(_FMT_STRING.format(arrow=_ASCII_ARROW))
    elif stream_format == LogFormats.JSON:
        stdout_formatter = _JsonFormatter()
    else:
        stdout_formatter = logging.Formatter(stream_format)

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.addFilter(_RangeFilter(0, 100).filter)
    stdout_handler.setFormatter(stdout_formatter)

    for handler in root_logger.handlers:
        root_logger.removeHandler(handler)  # Remove any existing handlers
    root_logger.addHandler(stdout_handler)

    return root_logger


def get_model_logger(name: str) -> ModelLogger:
    """Retrieve a ModelLogger with the specified name"""
    logging.setLoggerClass(ModelLogger)
    return logging.getLogger(name)  # Will return a ModelLogger based on the previous line


@contextmanager
def log_to_file(
    file_name: Union[str, PathLike],
    name: str,
    *,
    append: bool = False,
    raw_ascii: bool = False,
) -> Generator:
    """Context manager for opening and closing a logfile. Cleans up its file handler on exit.

    This is especially important during batch runs, because loggers are module-based (e.g. global). Without the cleanup,
    old file handlers would stick around and get written to.

    Args:
        file_name (str | PathLike): The filepath of the log file to write to
        name (str): The name of the logger to write log records from
        append (bool, optional): Defaults to ``False``. Option to append new log records to an existing log file
        raw_ascii (bool, optional): Defaults to ``False``. Ensures log file only contains valid ASCII characters
    """
    root = logging.getLogger(name)

    write_mode = "a" if append else "w"
    handler = FileHandler(file_name, mode=write_mode, encoding="utf-8")

    arrow = _ASCII_ARROW if raw_ascii else _UNC_ARROW
    fmt_str = _FMT_STRING.format(arrow=arrow)
    handler.setFormatter(Formatter(fmt_str))

    handler.addFilter(_RangeFilter(0, 100).filter)

    root.addHandler(handler)
    try:
        yield
    except:
        with open(file_name, mode="a") as writer:
            writer.write(str("\n" + "-" * 100 + "\n\n"))
            writer.write(tb.format_exc())
        raise
    finally:
        root.removeHandler(handler)


# endregion
